from django.utils.translation import ugettext_lazy as _

__title__ = 'dash.constants'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'ACTION_CHOICE_APPEND',
    'ACTION_CHOICE_REPLACE',
    'ACTION_CHOICES',
)


ACTION_CHOICE_REPLACE = '1'
ACTION_CHOICE_APPEND = '2'
ACTION_CHOICES = (
    (ACTION_CHOICE_APPEND, _("Append")),
    (ACTION_CHOICE_REPLACE, _("Replace")),
)



from django.utils.translation import ugettext_lazy as _

from vishap import render_video

from ....base import BaseDashboardPlugin
from ....factory import plugin_factory

from .forms import VideoForm

__title__ = 'dash.contrib.plugins.video.dash_plugins'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = ('BaseVideoPlugin',)


# ***************************************************************************
# ******************************* Base Video plugin *************************
# ***************************************************************************


class BaseVideoPlugin(BaseDashboardPlugin):
    """Base Video plugin."""

    name = _("Video")
    group = _("Internet")
    form = VideoForm
    html_classes = ['video']

    def post_processor(self):
        """Post process."""
        self.data.embed_code = render_video(self.data.url)

# ***************************************************************************
# ********** Generating and registering the plugins using factory ***********
# ***************************************************************************


sizes = (
    (1, 1),
    (2, 2),
    (3, 3),
    (4, 4),
    (5, 5),
    (6, 6)
)

plugin_factory(BaseVideoPlugin, 'video', sizes)



import logging

from django.utils.translation import ugettext_lazy as _
from django.core.cache import cache

from six.moves.urllib.request import urlopen

from ....base import BaseDashboardPlugin
from ....factory import plugin_factory
from ....json_package import json
from ....settings import DEBUG

from .forms import WeatherForm
from .settings import API_KEY, API_ENDPOINT_URL

__title__ = 'dash.contrib.plugins.weather.dash_plugins'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = ('BaseWeatherPlugin',)

logger = logging.getLogger(__name__)

# ***************************************************************************
# ************************** Base Weather plugin ****************************
# ***************************************************************************


class BaseWeatherPlugin(BaseDashboardPlugin):
    """Base Weather plugin."""

    name = _("Weather")
    form = WeatherForm
    group = _("Weather")

    def post_processor(self):
        """Post process.

        If no text available, use dummy.
        """
        key = '{0}-{1}-{2}'.format(self.layout_uid,
                                   self.placeholder_uid,
                                   self.uid)
        self.data.weather_data_json = cache.get(key)

        if not self.data.weather_data_json:

            if self.data.public_ip:
                api_endpoint_url = API_ENDPOINT_URL.format(
                    API_KEY,
                    'json',
                    self.data.public_ip
                )

                try:
                    data = str(urlopen(api_endpoint_url).read())

                    self.data.weather_data_json = json.loads(data)

                    cache.set(
                        key,
                        self.data.weather_data_json,
                        int(self.data.cache_for)
                    )
                except Exception as err:
                    if DEBUG:
                        logger.debug(err)

        if self.data.weather_data_json:
            data = self.data.weather_data_json['data']

            try:
                current_condition = data['current_condition'][0]
            except (KeyError, TypeError, IndexError):
                current_condition = None

            if current_condition:
                self.data.current_cloudcover = current_condition['cloudcover']
                self.data.current_humidity = current_condition['humidity']
                self.data.current_pressure = current_condition['pressure']
                self.data.current_visibility = current_condition['visibility']
                self.data.current_temp_c = current_condition['temp_C']

                try:
                    self.data.current_weather_desc = \
                        current_condition['weatherDesc'][0]['value']
                except (KeyError, TypeError, IndexError):
                    pass

                try:
                    self.data.current_weather_icon_url = \
                        current_condition['weatherIconUrl'][0]['value']
                except (KeyError, TypeError, IndexError):
                    pass

            try:
                weather = data['weather'][0]
            except (KeyError, TypeError, IndexError):
                weather = None

            if weather:
                self.data.temp_max_c = weather['tempMaxC']
                self.data.temp_min_c = weather['tempMinC']
                self.data.windspeed_kmph = weather['windspeedKmph']
                self.data.wind_dir_16_point = weather['winddir16Point']

                try:
                    self.data.weather_desc = weather['weatherDesc'][0]['value']
                except (KeyError, TypeError, IndexError):
                    pass

                try:
                    self.data.weather_icon_url = \
                        weather['weatherIconUrl'][0]['value']
                except (KeyError, TypeError, IndexError):
                    pass


# ****************************************************************************
# ********** Generating and registering the plugins using factory ************
# ****************************************************************************


sizes = (
    (2, 2),
    (3, 3),
)

plugin_factory(BaseWeatherPlugin, 'weather', sizes)



from django.utils.translation import ugettext_lazy as _

__title__ = 'dash.contrib.layouts.bootstrap3.defaults'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'IMAGE_CHOICES',
    'IMAGE_CHOICES_WITH_EMPTY_OPTION',
)


IMAGE_CHOICES = (
    # Icons that are also present in `URLPlugin`.
    ('glyphicon-film', _("Film")),
    # ('glyphicon-coffee', _("Coffee")),
    ('glyphicon-calendar', _("Calendar")),
    ('glyphicon-book', _("Book")),
    ('glyphicon-music', _("Music")),
    ('glyphicon-picture', _("Picture")),
    # ('glyphicon-rss-sign', _("RSS")),
    ('glyphicon-star', _("Star")),
    ('glyphicon-thumbs-up', _("Thumbs-up")),
    # ('glyphicon-smile', _("Smile")),
    # ('glyphicon-gamepad', _("Gamepad")),
    ('glyphicon-plane', _("Plane")),
    ('glyphicon-road', _("Road")),
    ('glyphicon-camera', _("Camera")),
    ('glyphicon-download', _("Download")),
    # ('glyphicon-food', _("Food")),
    ('glyphicon-info-sign', _("Info")),
    ('glyphicon-shopping-cart', _("Shopping cart")),
    # ('glyphicon-truck', _("Truck")),
    ('glyphicon-wrench', _("Wrench")),
    # ('glyphicon-facebook', _("Facebook")),
    # ('glyphicon-github', _("Github")),
    # ('glyphicon-google-plus', _("Google plus")),
    # ('glyphicon-linkedin', _("LinkedIn")),
    # ('glyphicon-pinterest', _("Pinterest")),
    # ('glyphicon-twitter', _("Twitter")),
    # ('glyphicon-youtube', _("Youtube")),
    # ('glyphicon-bitbucket', _("Bitbucket")),
    # ('glyphicon-android', _("Android")),
    # ('glyphicon-apple', _("Apple")),
    # ('glyphicon-windows', _("Windows")),
    # ('glyphicon-tumblr-sign', _("Tumblr")),
    # ('glyphicon-instagram', _("Instagram")),
    # ('glyphicon-dropbox', _("Dropbox")),
    # ('glyphicon-trophy', _("Trophy")),
    # ('glyphicon-legal', _("Legal")),
    ('glyphicon-lock', _("Lock")),
    ('glyphicon-heart', _("Heart")),
    ('glyphicon-question-sign', _("Question")),
    ('glyphicon-headphones', _("Headphones")),
    ('glyphicon-gift', _("Gift")),
    # ('glyphicon-key', _("Key")),
    # ('glyphicon-female', _("Female")),
    # ('glyphicon-male', _("Male")),
    ('glyphicon-comment', _("Comment")),
    # ('glyphicon-bug', _("Bug")),
    ('glyphicon-bell', _("Bell")),
    ('glyphicon-search', _("Search")),
    ('glyphicon-map-marker', _("Map marker")),
    ('glyphicon-globe', _("Globe")),
    ('glyphicon-pencil', _("Pensil")),
    ('glyphicon-tasks', _("Tasks")),
)

IMAGE_CHOICES_WITH_EMPTY_OPTION = [('', '---------')] + list(IMAGE_CHOICES)



from django import forms
from django.utils.translation import ugettext_lazy as _

from .base import get_registered_layouts
from .constants import ACTION_CHOICES
from .models import DashboardWorkspace, DashboardSettings, DashboardPlugin

__title__ = 'dash.forms'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'BulkChangeDashboardPluginsForm',
    'DashboardSettingsForm',
    'DashboardWorkspaceForm',
)


class DashboardWorkspaceForm(forms.ModelForm):
    """Dashboard workspace form."""

    layout_uid = forms.TypedChoiceField(
        label=_('Layout'),
        choices=get_registered_layouts(),
        empty_value=None,
    )

    class Meta(object):
        """Meta."""

        model = DashboardWorkspace
        fields = ('layout_uid', 'user', 'name', 'is_public', 'is_clonable')

    def __init__(self, *args, **kwargs):
        different_layouts = kwargs.pop('different_layouts', False)
        super(DashboardWorkspaceForm, self).__init__(*args, **kwargs)
        self.fields['user'].widget = forms.widgets.HiddenInput()

        if not different_layouts:
            self.fields['layout_uid'].widget = forms.widgets.HiddenInput()


class DashboardSettingsForm(forms.ModelForm):
    """Dashboard settings form."""

    class Meta(object):
        """Meta."""

        model = DashboardSettings
        fields = ('user', 'layout_uid', 'title', 'is_public')

    layout_uid = forms.ChoiceField(choices=get_registered_layouts())

    def __init__(self, *args, **kwargs):
        super(DashboardSettingsForm, self).__init__(*args, **kwargs)
        self.fields['user'].widget = forms.widgets.HiddenInput()


class BulkChangeDashboardPluginsForm(forms.ModelForm):
    """Bulk change dashboard plugins form.

    - `selected_dashboard_plugins` (str): List of comma separated values to be
      changed.
    - `users_action` (int): For indicating whether the users shall be
       appended to the dashboard plugins or replaced.
    - `groups_action` (int): For indicating whether the groups shall be
      appended to the dashboard plugins or replaced.
    """
    selected_dashboard_plugins = forms.CharField(
        required=True,
        label=_("Selected dashboard plugins"),
        widget=forms.widgets.HiddenInput
    )
    users_action = forms.ChoiceField(
        required=False,
        label=_("Users action"),
        choices=ACTION_CHOICES,
        help_text=_("If set to ``replace``, the groups are replaced; "
                    "otherwise - appended.")
    )
    groups_action = forms.ChoiceField(
        required=False,
        label=_("Groups action"),
        choices=ACTION_CHOICES,
        help_text=_("If set to ``replace``, the groups are replaced; "
                    "otherwise - appended.")
    )

    class Meta(object):
        """Meta."""

        model = DashboardPlugin
        fields = ('groups', 'groups_action', 'users', 'users_action',)

    class Media(object):
        """Media."""

        css = {
            'all': ('css/admin_custom.css',)
        }

    def __init__(self, *args, **kwargs):
        super(BulkChangeDashboardPluginsForm, self).__init__(*args, **kwargs)
        self.fields['users'].required = False
        self.fields['groups'].required = False



from ....base import (
    BaseDashboardLayout,
    BaseDashboardPlaceholder,
    layout_registry,
)

__title__ = 'dash.contrib.layouts.windows8.dash_layouts'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = ('Windows8Layout',)


class Windows8MainPlaceholder(BaseDashboardPlaceholder):
    """Main placeholder."""

    uid = 'main'
    cols = 6
    rows = 4
    cell_width = 140
    cell_height = 135


class WindowsSidebarPlaceholder(BaseDashboardPlaceholder):
    """Sidebar placeholder."""

    uid = 'sidebar'
    cols = 2
    rows = 4
    cell_width = 140
    cell_height = 135


class Windows8Layout(BaseDashboardLayout):
    """Windows8 layout."""

    uid = 'windows8'
    name = 'Windows 8'
    view_template_name = 'windows8/view_layout.html'
    edit_template_name = 'windows8/edit_layout.html'
    placeholders = [Windows8MainPlaceholder, WindowsSidebarPlaceholder]
    cell_units = 'px'
    media_css = (
        'windows8/css/dash_solid_borders.css',
        'windows8/css/dash_layout_windows8.css',
    )
    # media_js = ('js/dash_layout_windows8.js',)


layout_registry.register(Windows8Layout)



from django.template.loader import render_to_string

from ....base import BaseDashboardPluginWidget

__title__ = 'dash.contrib.plugins.weather.dash_widgets'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'BaseWeatherWidget',
    'Weather2x2Widget',
    'Weather3x3Widget',
)

# **********************************************************************
# *********************** Base weather widget plugin *******************
# **********************************************************************


class BaseWeatherWidget(BaseDashboardPluginWidget):
    """Base weather plugin widget."""

    # media_js = [
    #    'js/dash_plugin_weather.js',
    # ]
    media_css = [
        'css/dash_plugin_weather.css',
    ]

    def render(self, request=None):
        """Render."""
        context = {'plugin': self.plugin}
        return render_to_string('weather/render.html', context)

# **********************************************************************
# ************************** Specific widgets **************************
# **********************************************************************


class Weather2x2Widget(BaseWeatherWidget):
    """Weather plugin 2x2 widget."""

    plugin_uid = 'weather_2x2'
    cols = 2
    rows = 2


class Weather3x3Widget(BaseWeatherWidget):
    """Weather plugin 3x3 widget."""

    plugin_uid = 'weather_3x3'
    cols = 3
    rows = 3



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import autoslug.fields
import dash.helpers


class Migration(migrations.Migration):

    dependencies = [
        ('dash', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='dashboardsettings',
            name='allow_different_layouts',
            field=models.BooleanField(default=False, help_text='Allows you to use different layouts for each workspace.', verbose_name='Allow different layouts per workspace?'),
        ),
        migrations.AlterField(
            model_name='dashboardworkspace',
            name='slug',
            field=autoslug.fields.AutoSlugField(editable=False, populate_from=b'name', unique=True, verbose_name='Slug', slugify=dash.helpers.slugify_workspace),
        ),
    ]



import glob
import logging
import os
import shutil
import uuid

from django.conf import settings
from django.core.files.base import File

from .settings import (
    FIT_METHOD_CROP_CENTER,
    FIT_METHOD_CROP_SCALE,
    FIT_METHOD_CROP_SMART,
    IMAGES_UPLOAD_DIR,
)

__title__ = 'dash.contrib.plugins.image.helpers'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'clone_file',
    'delete_file',
    'ensure_unique_filename',
    'get_crop_filter',
    'handle_uploaded_file',
)


logger = logging.getLogger(__file__)

IMAGES_UPLOAD_DIR_ABSOLUTE_PATH = os.path.join(settings.MEDIA_ROOT,
                                               IMAGES_UPLOAD_DIR)
if not os.path.exists(IMAGES_UPLOAD_DIR_ABSOLUTE_PATH):
    os.makedirs(IMAGES_UPLOAD_DIR_ABSOLUTE_PATH)


def ensure_unique_filename(destination):
    """Ensure unique filename.

    Makes sure filenames are never overwritten. If file name already exists,
    makes a new one based on the first 50 chars of the original file name,
    having a uuid4 appended afterwards.

    :param string destination:
    :return string:
    """
    if os.path.exists(destination):
        dest, filename = os.path.split(destination)
        filename, extension = os.path.splitext(filename)
        return os.path.join(
            dest, "{0}_{1}{2}".format(filename[:30], uuid.uuid4(), extension)
        )
    else:
        return destination


def handle_uploaded_file(image_file):
    """Handle uploaded file.

    :param django.core.files.uploadedfile.InMemoryUploadedFile image_file:
    :return string: Path to the image (relative).
    """
    if isinstance(image_file, File):
        destination_path = ensure_unique_filename(
            os.path.join(IMAGES_UPLOAD_DIR_ABSOLUTE_PATH, image_file.name)
        )
        image_filename = image_file.name
        with open(destination_path, 'wb+') as destination:
            image_filename = os.path.basename(destination.name)
            for chunk in image_file.chunks():
                destination.write(chunk)
        return os.path.join(IMAGES_UPLOAD_DIR, image_filename)
    return image_file


def get_crop_filter(fit_method):
    """Get crop filter."""
    if fit_method in (FIT_METHOD_CROP_SMART,
                      FIT_METHOD_CROP_CENTER,
                      FIT_METHOD_CROP_SCALE):
        return fit_method


def delete_file(image_file):
    """Delete file from disc."""
    try:
        # Delete the main file.
        file_path = os.path.join(settings.MEDIA_ROOT, image_file)
        os.remove(file_path)

        # Delete the sized version of it.
        files = glob.glob("{0}*".format(file_path))
        for _file in files:
            try:
                os.remove(_file)
            except Exception as err:
                logger.debug(str(err))

        # If all goes well...
        return True
    except Exception as err:
        logger.debug(str(err))
        return False


def clone_file(source_filename, relative_path=True):
    """Clone the file.

    :param string source_filename: Source filename.
    :param str relative_path:
    :return string: Filename of the cloned file.
    """
    if source_filename.startswith(IMAGES_UPLOAD_DIR):
        source_filename = os.path.join(settings.MEDIA_ROOT, source_filename)

    destination_filename = ensure_unique_filename(source_filename)
    try:
        shutil.copyfile(source_filename, destination_filename)
        if relative_path:
            destination_filename = destination_filename.replace(
                settings.MEDIA_ROOT, ''
            )
            if destination_filename.startswith('/'):
                destination_filename = destination_filename[1:]
        return destination_filename
    except Exception as err:
        logger.debug(str(err))



from django.core.management.base import BaseCommand

from ...utils import update_plugin_data_for_entries

__title__ = 'dash.management.commands.dash_update_plugin_data'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = ('Command',)


class Command(BaseCommand):
    """Update the plugin data for all dashboard entries of all users.

    Rules for update are specified in the plugin itself.

    This command shall be ran if significant changes have been made to the
    system for which the data shall be updated.
    """

    def handle(self, *args, **options):

        update_plugin_data_for_entries()



from autoslug.settings import slugify

from django.utils.encoding import force_text

from six import PY3

__title__ = 'dash.helpers'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'clean_plugin_data',
    'clone_plugin_data',
    'iterable_to_dict',
    'lists_overlap',
    'safe_text',
    'slugify_workspace',
    'uniquify_sequence',
    'update_plugin_data',
)


def slugify_workspace(val):
    """Slugify workspace."""
    return slugify(val.lower()).lower()


def safe_text(text):
    """Safe text (encode).

    :return str:
    """
    if PY3:
        return force_text(text, encoding='utf-8')
    else:
        return force_text(text, encoding='utf-8').encode('utf-8')


def lists_overlap(sub, main):
    for i in sub:
        if i in main:
            return True
    return False


def iterable_to_dict(items, key_attr_name):
    """Convert iterable of certain objects to dict.

    :param iterable items:
    :param string key_attr_name: Attribute to use as a dictionary key.
    :return dict:
    """
    items_dict = {}
    for item in items:
        items_dict.update({getattr(item, key_attr_name): item})
    return items_dict


def uniquify_sequence(sequence):
    """Makes the sequence unique.

    Makes sure items in the given sequence are unique, having the original
    order preserved.

    :param iterable sequence:
    :return list:
    """
    seen = set()
    seen_add = seen.add
    return [x for x in sequence if x not in seen and not seen_add(x)]


def clean_plugin_data(dashboard_entries, request=None):
    """Clean up the plugin data.

    Clean up the plugin data (database, files) for the dashboard_entries given.

    :param iterable dashboard_entries:
    :param django.http.HttpRequest request:
    :return bool: Boolean True if no errors occurred and False otherwise.
    """
    errors = False
    for dashboard_entry in dashboard_entries:
        plugin = dashboard_entry.get_plugin(request=request)

        if plugin:
            plugin._delete_plugin_data()

    return not errors


def clone_plugin_data(dashboard_entry, request=None):
    """Clone plugin data of a dashboard entry."""
    if dashboard_entry:
        plugin = dashboard_entry.get_plugin(request=request)

        if plugin:
            plugin_data = plugin._clone_plugin_data(dashboard_entry)

            if plugin_data is None:
                plugin_data = dashboard_entry.plugin_data

            return plugin_data


def update_plugin_data(dashboard_entry, request=None):
    """Update plugin data of a dashboard entry."""
    if dashboard_entry:
        plugin = dashboard_entry.get_plugin(request=request)

        if plugin:
            return plugin._update_plugin_data(dashboard_entry)



from django.template.loader import render_to_string

from nine import versions

from ....base import BaseDashboardPluginWidget

if versions.DJANGO_GTE_1_8:
    from django.template.context_processors import csrf
else:
    from django.core.context_processors import csrf

__title__ = 'dash.contrib.plugins.rss_feed.dash_widgets'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'BaseReadRSSFeedWidget',
    'ReadRSSFeed2x3Widget',
    'ReadRSSFeed3x3Widget',
)

# ************************************************************************
# ****************** Android widgets for Read RSS feed plugin ************
# ************************************************************************


class BaseReadRSSFeedWidget(BaseDashboardPluginWidget):
    """Base read RSS feed plugin widget."""

    media_js = [
        'js/dash_plugin_read_rss_feed.js',
    ]
    media_css = [
        'css/dash_plugin_read_rss_feed.css',
    ]

    def render(self, request=None):
        """Render."""
        context = {
            'plugin': self.plugin,
            'csrfmiddlewaretoken': csrf(request)
        }
        return render_to_string('rss_feed/render.html', context)

# ***********************************************************************
# ********************** Specific widgets for Memo plugin ***************
# ***********************************************************************


class ReadRSSFeed2x3Widget(BaseReadRSSFeedWidget):
    """Read RSS feed 2x3 plugin widget."""

    plugin_uid = 'read_rss_feed_2x3'
    cols = 2
    rows = 3


class ReadRSSFeed3x3Widget(BaseReadRSSFeedWidget):
    """Big read RSS 3x3 feed plugin widget."""

    plugin_uid = 'read_rss_feed_3x3'
    cols = 3
    rows = 3



__title__ = 'dash.contrib.plugins.rss_feed.defaults'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'DEFAULT_CACHE_FOR',
    'DEFAULT_MAX_FEED_ITEMS',
    'DEFAULT_SHOW_TITLE',
    'DEFAULT_TRUNCATE_AFTER',
)

DEFAULT_CACHE_FOR = 3600
DEFAULT_MAX_FEED_ITEMS = 6
DEFAULT_SHOW_TITLE = True
DEFAULT_TRUNCATE_AFTER = 35



import copy
import datetime
import logging

from django.core.exceptions import ObjectDoesNotExist

from nine import versions

from .base import (
    BaseDashboardLayout,
    BaseDashboardPlaceholder,
    ensure_autodiscover,
    get_layout,
    get_registered_plugin_uids,
    get_registered_plugins,
    plugin_registry,
    plugin_widget_registry,
    PluginWidgetRegistry,
)
from .exceptions import PluginWidgetOutOfPlaceholderBoundaries
from .helpers import (
    clone_plugin_data,
    lists_overlap,
    safe_text,
    slugify_workspace,
    update_plugin_data,
)
from .models import (
    DashboardEntry,
    DashboardPlugin,
    DashboardSettings,
    DashboardWorkspace,
)
from .settings import DEBUG, RESTRICT_PLUGIN_ACCESS

if versions.DJANGO_GTE_1_10:
    from django.urls import reverse, NoReverseMatch
else:
    from django.core.urlresolvers import reverse, NoReverseMatch

__title__ = 'dash.utils'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'build_cells_matrix',
    'get_allowed_plugin_uids',
    'get_dashboard_settings',
    'get_or_create_dashboard_settings',
    'get_public_dashboard_url',
    'get_user_plugin_uids',
    'get_user_plugins',
    'get_widgets',
    'get_workspaces',
    'sync_plugins',
    'update_plugin_data_for_entries',
)


logger = logging.getLogger(__name__)


def get_allowed_plugin_uids(user):
    """Gets allowed plugins uids for user given.

    :param django.contrib.auth.models import User:
    :return list:
    """
    try:
        queryset_groups = DashboardPlugin \
            ._default_manager \
            .filter(groups__in=user.groups.all()) \
            .distinct()
        queryset_users = DashboardPlugin \
            ._default_manager \
            .filter(users=user)\
            .distinct()
        queryset = queryset_groups | queryset_users
        queryset = queryset.only('plugin_uid')
        return [obj.plugin_uid for obj in queryset]
    except Exception as err:
        if DEBUG:
            logger.debug(err)
        return []


def get_user_plugins(user):
    """Get a list of user plugins.

    Gets a list of user plugins in a form if tuple (plugin name, plugin
    description). If not yet auto-discovered, auto-discovers them.

    :return list:
    """
    ensure_autodiscover()

    if not RESTRICT_PLUGIN_ACCESS or getattr(user, 'is_superuser', False):
        return get_registered_plugins()

    registered_plugins = []

    allowed_plugin_uids = get_allowed_plugin_uids(user)

    for uid, plugin in plugin_registry._registry.items():
        if uid in allowed_plugin_uids:
            registered_plugins.append((uid, safe_text(plugin.name)))

    return registered_plugins


def get_user_plugin_uids(user):
    """Get a list of user plugin uids as a list .

    If not yet auto-discovered, auto-discovers them.

    :return list:
    """
    ensure_autodiscover()

    if not RESTRICT_PLUGIN_ACCESS or getattr(user, 'is_superuser', False):
        return get_registered_plugin_uids()

    registered_plugins = []

    allowed_plugin_uids = get_allowed_plugin_uids(user)

    for uid, plugin in plugin_registry._registry.items():
        if uid in allowed_plugin_uids:
            registered_plugins.append(uid)

    return registered_plugins


def get_widgets(layout, placeholder, user=None, workspace=None,
                position=None, occupied_cells=[], sort_items=True):
    """Get widgets.

    In case if in restricted mode (``RESTRICT_PLUGIN_ACCESS`` is set to True),
    user argument should be provided. Based on it, the list of plugins is
    returned.  Restrictions are bypassed in case if ``RESTRICT_PLUGIN_ACCESS``
    is set to False or user given is a superuser.

    Placeholders are validated already. We don't need to have validation here.

    :param dash.base.BaseDashLayout layout: Layout object.
    :param string placeholder: Placeholder uid.
    :param django.contrib.auth.models.User user:
    :param string workspace: Workspace slug.
    :param int position: Plugin position.
    :param list occupied_cells: List of already occupied cells.
    :param bool sort_items: If set to True, returned items are sorted.
    :return list:
    """
    # We should get the layout, see loop through its' plugins and see which of
    # those do have renderers. Then we get all the plugins (based on whether
    # they are restricted or not - get the list) and then filter out those
    # that do not have renderers.

    ensure_autodiscover()

    registered_widgets = {}

    plugin_widget_uids = plugin_widget_registry._registry.keys()

    if not RESTRICT_PLUGIN_ACCESS or getattr(user, 'is_superuser', False):
        for uid, plugin in plugin_registry._registry.items():
            # We should make sure that there are widgets available for the
            # placeholder.
            plugin_widget_uid = PluginWidgetRegistry.namify(
                layout.uid,
                placeholder.uid,
                uid
            )

            # Get cells occupied by plugin widget.
            widget_occupied_cells = get_occupied_cells(
                layout,
                placeholder,
                uid,
                position,
                check_boundaries=True
            )

            if plugin_widget_uid in plugin_widget_uids \
               and widget_occupied_cells is not False \
               and not lists_overlap(widget_occupied_cells, occupied_cells):

                plugin_widget = plugin_widget_registry.get(plugin_widget_uid)
                kwargs = {
                    'placeholder_uid': placeholder.uid,
                    'plugin_uid': uid
                }
                if workspace:
                    kwargs.update({'workspace': workspace})
                if position:
                    kwargs.update({'position': position})

                plugin_group = safe_text(plugin.group)
                if plugin_group not in registered_widgets:
                    registered_widgets[plugin_group] = []

                widget_name = safe_text(plugin.name)

                registered_widgets[plugin_group].append(
                    (
                        uid,
                        '{0} ({1}x{2})'.format(widget_name,
                                               plugin_widget.cols,
                                               plugin_widget.rows),
                        reverse('dash.add_dashboard_entry', kwargs=kwargs)
                    )
                )
    else:
        allowed_plugin_uids = get_allowed_plugin_uids(user)

        for uid, plugin in plugin_registry._registry.items():
            # We should make sure that there are widgets available for the
            # placeholder and user has access to the widget desired.
            plugin_widget_uid = PluginWidgetRegistry.namify(
                layout.uid,
                placeholder.uid,
                uid
            )

            # Get cells occupied by plugin widget.
            widget_occupied_cells = get_occupied_cells(
                layout,
                placeholder,
                uid,
                position,
                check_boundaries=True
            )

            if uid in allowed_plugin_uids \
               and plugin_widget_uid in plugin_widget_uids \
               and widget_occupied_cells is not False \
               and not lists_overlap(widget_occupied_cells, occupied_cells):

                plugin_widget = plugin_widget_registry.get(plugin_widget_uid)
                kwargs = {
                    'placeholder_uid': placeholder.uid,
                    'plugin_uid': uid
                }
                if workspace:
                    kwargs.update({'workspace': workspace})
                if position:
                    kwargs.update({'position': position})

                plugin_group = safe_text(plugin.group)
                if plugin_group not in registered_widgets:
                    registered_widgets[plugin_group] = []

                registered_widgets[plugin_group].append(
                    (
                        uid,
                        '{0} ({1}x{2})'.format(safe_text(plugin.name),
                                               plugin_widget.cols,
                                               plugin_widget.rows),
                        reverse('dash.add_dashboard_entry', kwargs=kwargs)
                    )
                )

    if sort_items:
        for key, prop in registered_widgets.items():
            prop.sort()

    return registered_widgets


def update_plugin_data_for_entries(dashboard_entries=None, request=None):
    """Update the plugin data for all dashboard entries of all users.

    Rules for update are specified in the plugin itself.

    :param iterable dashboard_entries: If given, is used to iterate through
        and update the plugin data. If left empty, all dashboard entries will
        be updated.
    :param django.http.HttpRequest request:
    """
    if dashboard_entries is None:
        dashboard_entries = DashboardEntry._default_manager.all()

    for dashboard_entry in dashboard_entries:
        update_plugin_data(dashboard_entry, request=request)


def sync_plugins():
    """Sync the registered plugin list with data in
    ``dash.models.DashboardPlugin``.
    """
    # If not in restricted mode, the quit.

    if not RESTRICT_PLUGIN_ACCESS:
        return

    registered_plugins = set(get_registered_plugin_uids())

    synced_plugins = set([obj.plugin_uid for obj in
                          DashboardPlugin._default_manager.only('plugin_uid')])

    non_synced_plugins = registered_plugins - synced_plugins

    if not non_synced_plugins:
        return

    buf = []

    for plugin_uid in non_synced_plugins:
        buf.append(DashboardPlugin(plugin_uid=plugin_uid))

    DashboardPlugin._default_manager.bulk_create(buf)


def get_workspaces(user, layout_uid=None, workspace=None, public=False,
                   different_layouts=False):
    """Gets previous, current, next and and a queryset of all workspaces.

    :param django.contrib.auth.models.User user:
    :param str layout_uid:
    :param string workspace:
    :param bool public:
    :param bool different_layouts:
    :return dict:
    """
    # We need to show workspaces
    q_kwargs = {'user': user}
    if not different_layouts:
        q_kwargs.update({'layout_uid': layout_uid})
    if public:
        q_kwargs.update({'is_public': public})

    workspaces = list(
        DashboardWorkspace._default_manager
                          .filter(**q_kwargs)
                          .only('id', 'name', 'slug')
                          .order_by('position')[:]
    )

    next = None
    previous = None
    current = None
    current_not_found = False

    if workspace:

        # Slugifying the workspace
        workspace_slug = slugify_workspace(workspace)
        num_workspaces = len(workspaces)

        for index, ws in enumerate(workspaces):
            if workspace_slug == ws.slug:
                current = ws

                if index == 0:
                    # No previous workspace (previous is default).
                    try:
                        next = workspaces[1]
                    except IndexError:
                        pass

                elif num_workspaces == index:
                    # No next workspace (next is default).
                    try:
                        previous = workspaces[index - 1]
                    except IndexError:
                        pass

                else:
                    # Getting previous and next workspaces.
                    try:
                        previous = workspaces[index - 1]
                    except IndexError:
                        pass

                    try:
                        next = workspaces[index + 1]
                    except IndexError:
                        pass

        if current is None:
            current_not_found = True

    else:
        try:
            previous = workspaces[-1]
        except IndexError:
            pass

        try:
            next = workspaces[0]
        except IndexError:
            pass

    return {
        'workspaces': workspaces,
        'next_workspace': next,
        'previous_workspace': previous,
        'current_workspace': current,
        'current_workspace_not_found': current_not_found
    }


def get_occupied_cells(layout, placeholder, plugin_uid, position,
                       check_boundaries=False, fail_silently=True):
    """
    Get cells occupied by the given dashboard entry.

    :param dash.base.BaseDashboardLayout layout: Instance of
        subclassed ``dash.base.BaseDashboardLayout`` object.
    :param dash.base.BaseDashboardPlaceholder placeholder: Instance of
        subclassed ``dash.base.BaseDashboardPlaceholder`` object.
    :param string plugin_uid: UID of the plugin to check against.
    :param int position: Position of the plugin to check against.
    :param bool check_boundaries: If set to True, boundaries of the
        placeholders are also considered.
    :param bool fail_silently: If set to True, no exceptions are raised.
    :return mixed: Returns a list (could be an empty list as well) if all
        goes well and returns boolean False if out of the placeholder
        boundaries.
    """
    assert isinstance(layout, BaseDashboardLayout)
    # assert issubclass(placeholder, BaseDashboardPlaceholder)

    widget_cls = plugin_widget_registry.get(
        PluginWidgetRegistry.namify(layout.uid, placeholder.uid, plugin_uid)
    )
    occupied_cells = []
    placeholder_max_cell_num = placeholder.cols * placeholder.rows

    try:
        position = int(position)
    except Exception as err:
        if fail_silently:
            return False
        else:
            raise err

    if widget_cls:

        # First check the basic things.
        if check_boundaries:
            # Checking if widget isn't touching the boundaries. Checking the
            # widget width.
            relative_col_num = position % placeholder.cols
            if 0 == relative_col_num:
                relative_col_num = placeholder.cols

            if (relative_col_num + widget_cls.cols - 1) > placeholder.cols:
                if fail_silently:
                    return False
                else:
                    raise PluginWidgetOutOfPlaceholderBoundaries(
                        "Widget is out of placeholder boundaries."
                    )

            # Checking if widget isn't touching the boundaries. Checking the
            # widget height.
            # relative_row_num = position % placeholder.rows
            # if 0 == relative_row_num:
            #     relative_row_num = placeholder.rows
            # if (relative_row_num + widget_cls.rows - 1) > placeholder.rows:
            #     if fail_silently:
            #         return False
            #     else:
            #         raise PluginWidgetOutOfPlaceholderBoundaries(
            #             "Widget is out of placeholder boundaries."
            #         )

        # Now check the collision with other plugin widgets.
        for row in range(0, widget_cls.rows):
            for col in range(0, widget_cls.cols):
                cell_num = position + col + (row * placeholder.cols)

                if check_boundaries and cell_num > placeholder_max_cell_num:
                    if fail_silently:
                        return False
                    else:
                        raise PluginWidgetOutOfPlaceholderBoundaries(
                            "Widget is out of placeholder boundaries."
                        )
                occupied_cells.append(cell_num)

    return occupied_cells


def build_cells_matrix(user, layout, placeholder, workspace=None):
    """Build the cells matrix.

    :param django.contrib.auth.models.User user:
    :param dash.base.BaseDashboardLayout layout: Instance of
        subclassed ``dash.base.BaseDashboardLayout`` object.
    :param dash.base.BaseDashboardPlaceholder placeholder: Instance of
        subclassed ``dash.base.BaseDashboardPlaceholder`` object.
    :param string workspace: Workspace slug.
    :return list: List of cells occupied.
    """
    assert isinstance(layout, BaseDashboardLayout)
    assert issubclass(placeholder, BaseDashboardPlaceholder)

    # Getting the list of plugins that user is allowed to use.
    registered_plugins = get_user_plugins(user)
    user_plugin_uids = [uid for uid, repr in registered_plugins]

    # Getting the queryset for user and freezing it.
    dashboard_entries = DashboardEntry \
        ._default_manager \
        .get_for_user(
            user=user,
            layout_uid=layout.uid,
            workspace=workspace) \
        .select_related('workspace', 'user') \
        .filter(plugin_uid__in=user_plugin_uids) \
        .filter(placeholder_uid=placeholder.uid) \
        .order_by('placeholder_uid', 'position')[:]

    matrix = []
    for dashboard_entry in dashboard_entries:
        occupied_cells = get_occupied_cells(
            layout,
            placeholder,
            dashboard_entry.plugin_uid,
            dashboard_entry.position
        )
        if occupied_cells:
            matrix += occupied_cells
        # Now we should calculate how much space each widget occupies.

    return matrix


def get_or_create_dashboard_settings(user):
    """
    Gets dashboard settings for the user given. If no settings found, creates
    default settings.

    :param django.contrib.auth.models.User user:
    :return dash.models.DashboardSettings: Returns
        ``dash.models.DashboardSettings`` instance.
    """
    # Check if user trying to edit the dashboard workspace actually owns it.
    try:
        dashboard_settings = DashboardSettings._default_manager \
                                              .select_related('user') \
                                              .get(user=user)
    except ObjectDoesNotExist as err:
        layout = get_layout(as_instance=True)
        dashboard_settings = DashboardSettings()
        dashboard_settings.layout_uid = layout.uid
        dashboard_settings.user = user
        dashboard_settings.save()

    return dashboard_settings


def get_dashboard_settings(username):
    """Get dashboard settings for the user given.

    If no settings found, creates default settings.

    :param string username:
    :return dash.models.DashboardSettings: Returns
        ``dash.models.DashboardSettings`` instance.
    """
    # Check if user trying to edit the dashboard workspace actually owns it.
    try:
        return DashboardSettings._default_manager \
                                .select_related('user') \
                                .get(user__username=username)
    except ObjectDoesNotExist:
        pass


def get_public_dashboard_url(dashboard_settings):
    """Get resolved public dashboard URL

    Get resolved public dashboard URL if public dashboard app is
    installed == present in the global urls module of the project).

    :param dash.models.DashboardSettings dashboard_settings: Instance of
        ``dash.models.DashboardSettings``.
    :return string:
    """
    if dashboard_settings.is_public:
        try:
            # Resolve URL
            return reverse(
                'dash.public_dashboard',
                kwargs={'username': dashboard_settings.user.username}
            )
        except NoReverseMatch:
            # Most likely, the public dashboard is not present
            pass
    return ''


def clone_workspace(workspace, for_user, request=None):
    """Clone entire workspace.

    :param dash.models.DashboardWorkspace workspace:
    :param django.contrib.auth.models.User for_user:
    :param django.http.HttpRequest request:
    :return dash.models.DashboardWorkspace: Cloned workspace instance.
    """
    # Cloning workspace object.
    cloned_workspace = copy.copy(workspace)
    cloned_workspace.pk = None
    cloned_workspace.user = for_user
    cloned_workspace.is_public = False
    cloned_workspace.is_clonable = False
    cloned_workspace.name = "{0} cloned on {1}".format(
        cloned_workspace.name,
        datetime.datetime.now()
    )

    cloned_workspace.save()

    # Cloning workspace entries.
    dashboard_entries = DashboardEntry._default_manager \
                                      .filter(workspace=workspace)

    buf = []

    for dashboard_entry in dashboard_entries:
        cloned_plugin_data = clone_plugin_data(dashboard_entry,
                                               request=request)
        cloned_dashboard_entry = DashboardEntry(
            user=for_user,
            workspace=cloned_workspace,
            layout_uid=dashboard_entry.layout_uid,
            placeholder_uid=dashboard_entry.placeholder_uid,
            plugin_uid=dashboard_entry.plugin_uid,
            plugin_data=cloned_plugin_data,
            position=dashboard_entry.position,
        )

        buf.append(cloned_dashboard_entry)

    DashboardEntry._default_manager.bulk_create(buf)

    return cloned_workspace



import sys
import os
from setuptools import setup, find_packages

version = '0.5'

try:
    readme = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()
    screen_shots = open(
        os.path.join(os.path.dirname(__file__), 'SCREENSHOTS.rst')
    ).read()
    screen_shots = screen_shots.replace(
        '.. image:: _static',
        '.. figure:: https://github.com/barseghyanartur/django-dash/raw/'
        'master/docs/_static'
    )
except Exception:
    readme = ''
    screen_shots = ''

template_dirs = [
    "src/dash/templates/dash",  # Core templates
    # Android layout
    "src/dash/contrib/layouts/android/templates/android",
    # Bootstrap 2 layouts
    "src/dash/contrib/layouts/bootstrap2/templates/bootstrap2",
    # Bootstrap 3 layouts
    "src/dash/contrib/layouts/bootstrap3/templates/bootstrap3",
    # Windows 8 layout
    "src/dash/contrib/layouts/windows8/templates/windows8",
    "src/dash/contrib/plugins/dummy/templates/dummy",  # Dummy plugin
    "src/dash/contrib/plugins/image/templates/image",  # Image plugin
    "src/dash/contrib/plugins/memo/templates/memo",  # Memo plugin
    # "src/dash/contrib/plugins/news/templates/news",  # News plugin
    "src/dash/contrib/plugins/rss_feed/templates/rss_feed",  # RSS feed plugin
    "src/dash/contrib/plugins/url/templates/url",  # URL plugin
    "src/dash/contrib/plugins/video/templates/video",  # Video plugin
    "src/dash/contrib/plugins/weather/templates/weather",  # Weather plugin
]
static_dirs = [
    "src/dash/static", # Core static
    "src/dash/contrib/layouts/android/static",  # Android layout
    "src/dash/contrib/layouts/bootstrap2/static",  # Bootstrap 2 layouts
    # "src/dash/contrib/layouts/bootstrap3/static",  # Bootstrap 3 layouts
    "src/dash/contrib/layouts/windows8/static",  # Windows 8 layout
    "src/dash/contrib/plugins/image/static",  # Image plugin
    # "src/dash/contrib/plugins/news/static",  # News plugin
    "src/dash/contrib/plugins/rss_feed/static",  # RSS feed plugin
    "src/dash/contrib/plugins/url/static",  # URL plugin
    "src/dash/contrib/plugins/video/static",  # Video plugin
    "src/dash/contrib/plugins/weather/static",  # Weather plugin
]

locale_dirs = [
    "src/dash/locale/nl",
    "src/dash/locale/ru",
]

templates = []
static_files = []
locale_files = []

for template_dir in template_dirs:
    templates += [os.path.join(template_dir, f)
                  for f in os.listdir(template_dir)]

for static_dir in static_dirs:
    static_files += [os.path.join(static_dir, f)
                     for f in os.listdir(static_dir)]

for locale_dir in locale_dirs:
    locale_files += [os.path.join(locale_dir, f)
                     for f in os.listdir(locale_dir)]

install_requires = [
    'Pillow>=2.1.0',
    'django-autoslug>=1.7.1',
    'django-tinymce>=1.5.3',
    'feedparser>=5.1.3',
    'ordereddict>=1.1',
    'pif>=0.5,<1.0',
    'six>=1.9',
    # 'easy-thumbnails>=1.4,<2.0',
    'easy-thumbnails>=1.4',
    'transliterate>=1.5,<2.0',
    'vishap>=0.1.3,<2.0',
    'django-nine>=0.1.1',
    'Unidecode',
]

tests_require = [
    'radar>=0.3,<1.0',
    'selenium',
]

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

setup(
    name='django-dash',
    version=version,
    description="Customisable, modular dashboard application framework "
                "for Django.",
    long_description="{0}{1}".format(readme, screen_shots),
    classifiers=[
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Environment :: Web Environment",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "License :: OSI Approved :: GNU Lesser General Public License v2 "
        "or later (LGPLv2+)",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
    ],
    keywords='dashboard, django, django dashboard',
    author='Artur Barseghyan',
    author_email='artur.barseghyan@gmail.com',
    url='https://github.com/barseghyanartur/django-dash/',
    package_dir={'': 'src'},
    packages=find_packages(where='./src'),
    license='GPL 2.0/LGPL 2.1',
    install_requires=install_requires,
    tests_require=tests_require,
    package_data={
        'dash': templates + static_files + locale_files
    },
    include_package_data=True,
)



"""
All `uids` are supposed to be pythonic function names (see
PEP http://www.python.org/dev/peps/pep-0008/#function-names).
"""
import copy
import logging
import uuid

from django.forms import ModelForm
from django.http import Http404
from django.template.loader import render_to_string
from django.utils.translation import ugettext_lazy as _

from nine.versions import DJANGO_GTE_1_8

from .discover import autodiscover
from .exceptions import LayoutDoesNotExist, InvalidRegistryItemType
from .helpers import iterable_to_dict, uniquify_sequence, safe_text
from .json_package import json
from .settings import (
    ACTIVE_LAYOUT,
    DEBUG,
    DEFAULT_PLACEHOLDER_EDIT_TEMPLATE_NAME,
    DEFAULT_PLACEHOLDER_VIEW_TEMPLATE_NAME,
    LAYOUT_CELL_UNITS,
)


if DJANGO_GTE_1_8:
    from django.forms.utils import ErrorList
else:
    from django.forms.util import ErrorList

logger = logging.getLogger(__name__)

# _ = lambda s: s

__title__ = 'dash.base'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'BaseDashboardLayout',
    'BaseDashboardPlaceholder',
    'BaseDashboardPlugin',
    'BaseDashboardPluginWidget',
    'collect_widget_media',
    'DashboardPluginFormBase',
    'ensure_autodiscover',
    'get_layout',
    'get_registered_layout_uids',
    'get_registered_layouts',
    'get_registered_plugin_uids',
    'get_registered_plugins',
    'layout_registry',
    'plugin_registry',
    'plugin_widget_registry',
    'validate_placeholder_uid',
    'validate_plugin_uid',
)


class DashboardPluginFormBase(object):
    """Not a form actually. Defined for magic only.

    :property iterable plugin_data_fields: Fields to get when calling the
        ``get_plugin_data`` method. These field will be JSON serialized. All
        other fields, even if they are part of the form, won't be. Make sure
        all fields are serializable. If some of them aren't, override the
        ``save_plugin_data`` method and make them serializable there. See
        ``dash.contrib.plugins.image.forms`` as a good example.

    :example:

        >>> plugin_data_fields = (
        >>>    ('name', ''),
        >>>    ('active': False)
        >>> )
    """

    plugin_data_fields = None

    def _get_plugin_data(self, fields, request=None):
        """Gets plugin data.

        :param iterable fields: List of tuples to iterate.
        :param django.http.HttpRequest request:
        :return string: JSON dumpled string.
        """
        data = {}

        for field, default_value in fields:
            data.update({field: self.cleaned_data.get(field)})
        return json.dumps(data)

    def get_plugin_data(self, request=None):
        """Data that would be saved in the ``plugin_data`` field of the
        ``dash.models.DashboardEntry`` subclassed model.

        :param django.http.HttpRequest request:
        """
        if self.plugin_data_fields:
            return self._get_plugin_data(self.plugin_data_fields,
                                         request=request)

    def save_plugin_data(self, request=None):
        """Dummy, but necessary."""


class BaseDashboardLayout(object):
    """Base layout.

    Layouts consist of placeholders.

    :Properties:
        - `uid` (string): Layout unique identifier (globally).
        - `name` (string): Layout name.
        - `description` (string): Layout description.
        - `placeholders` (iterable): Iterable (list, tuple or set)
          of ``dash.base.BaseDashboardPlaceholder` subclasses.
        - `view_template_name` (string): Template used to render the
          layout (view).
        - `edit_template_name` (string): Template used to render the
          layout (edit).
        - `plugin_widgets_template_name_ajax` (string): Template used to
          render the plugin widgets popup.
        - `form_snippet_template_name` (string): Template used to render the
          forms.
        - `html_classes` (string): Extra HTML class that layout should get.
        - `cell_units` (string):
        - `media_css` (list): List all specific stylesheets.
        - `media_js` (list): List all specific javascripts.
    """

    uid = None
    name = None
    description = None
    placeholders = []
    view_template_name = None
    view_template_name_ajax = None
    edit_template_name = None
    edit_template_name_ajax = None
    plugin_widgets_template_name_ajax = 'dash/plugin_widgets_ajax.html'
    form_snippet_template_name = 'dash/snippets/generic_form_snippet.html'

    # 'dash/add_dashboard_entry_ajax.html'
    add_dashboard_entry_template_name = None

    # 'dash/add_dashboard_entry_ajax.html'
    add_dashboard_entry_ajax_template_name = None

    # 'dash/edit_dashboard_entry_ajax.html'
    edit_dashboard_entry_template_name = None

    # 'dash/edit_dashboard_entry_ajax.html'
    edit_dashboard_entry_ajax_template_name = None

    # 'dash/create_dashboard_workspace.html'
    create_dashboard_workspace_template_name = None

    # 'dash/create_dashboard_workspace_ajax.html'
    create_dashboard_workspace_ajax_template_name = None

    # 'dash/edit_dashboard_workspace.html'
    edit_dashboard_workspace_template_name = None

    # 'dash/edit_dashboard_workspace_ajax.html'
    edit_dashboard_workspace_ajax_template_name = None
    html_classes = []

    # Most likely, it makes sense to define this on a layout level.
    # Think of it.
    cell_units = None

    media_css = []
    media_js = []

    def __init__(self, user=None):
        """Constructor.

        :param django.contrib.auth.models.User user:
        """
        assert self.uid
        assert self.name
        assert self.view_template_name
        assert self.edit_template_name
        assert self.placeholders
        assert self.cell_units and self.cell_units in LAYOUT_CELL_UNITS
        assert isinstance(self.media_js, (list, tuple))
        assert isinstance(self.media_css, (list, tuple))

        if isinstance(self.media_js, tuple):
            self.media_js = list(self.media_js)

        if isinstance(self.media_css, tuple):
            self.media_css = list(self.media_css)

        self.placeholders_dict = {}
        self.placeholder_uids = []
        for placeholder in self.placeholders:
            self.placeholders_dict.update({placeholder.uid: placeholder})
            self.placeholder_uids.append(placeholder.uid)

        self.user = user
        self.widget_media_js = []
        self.widget_media_css = []

    def get_view_template_name(self, request=None, origin=None):
        """Get the view template name.

        :param django.http.HttpRequest request:
        :param string origin: Origin of the request. Hook to provide custom
            templates for apps. Example value: 'public_dashboard'. Take the
            `public_dashboard` app as example.
        """
        if not self.view_template_name_ajax:
            return self.view_template_name
        elif request and request.is_ajax():
            return self.view_template_name_ajax
        else:
            return self.view_template_name

    def get_edit_template_name(self, request=None):
        if not self.edit_template_name_ajax:
            return self.edit_template_name
        elif request and request.is_ajax():
            return self.edit_template_name_ajax
        else:
            return self.edit_template_name

    def get_placeholder(self, uid, default=None):
        return self.placeholders_dict.get(uid, default)

    def get_placeholders(self, request=None):
        """Get the list of placeholders registered for the layout.

        :param django.http.HttpRequest request:
        :return iterable: List of placeholder classes. Override in your layout
            if you need a custom behaviour.
        """
        return self.placeholders

    def get_placeholder_uids(self, request=None):
        """Get the list of placeholder uids.

        :param django.http.HttpRequest request:
        :return list:
        """
        uids = []
        for placeholder in self.placeholders:
            uids.append(placeholder.uid)
        return uids

    def collect_widget_media(self, dashboard_entries):
        """Collect the widget media files.

        :param iterable dashboard_entries: Iterable of
            ``dash.models.DashboardEntry`` instances.
        :return list:
        """
        widget_media = collect_widget_media(dashboard_entries)

        if widget_media:
            self.widget_media_js = widget_media['js']
            self.widget_media_css = widget_media['css']

    def get_media_css(self):
        """Get all CSS media files (for the layout + plugins).

        :return list:
        """
        media_css = self.media_css[:]
        if self.widget_media_css:
            media_css += self.widget_media_css

        media_css = uniquify_sequence(media_css)

        return media_css

    def get_media_js(self):
        """Get all JavaScript media files (for the layout + plugins).

        :return list:
        """
        media_js = self.media_js[:]
        if self.widget_media_js:
            media_js += self.widget_media_js

        media_js = uniquify_sequence(media_js)

        return media_js

    def get_grouped_dashboard_entries(self, dashboard_entries):
        """Get dashboard entries grouped by placeholder.

        :param iterable dashboard_entries: Iterable of
            ``dash.models.DashboardEntry`` objects.
        :return list:
        """
        entries = {}

        if not dashboard_entries:
            return entries

        for dashboard_entry in dashboard_entries:
            if dashboard_entry.placeholder_uid not in entries:
                entries[dashboard_entry.placeholder_uid] = []
            entries[dashboard_entry.placeholder_uid].append(dashboard_entry)

        return entries

    def get_placeholder_instances(self, dashboard_entries=None,
                                  workspace=None, request=None):
        """Get placeholder instances.

        :param iterable dashboard_entries: Iterable of
            ``dash.models.DashboardEntry`` objects.
        :param str workspace:
        :param django.http.HttpRequest request:
        :return list: List of `dash.base.BaseDashboardPlaceholder` subclassed
            instances.
        """
        entries = self.get_grouped_dashboard_entries(dashboard_entries)

        placeholder_instances = []

        for placeholder_cls in self.get_placeholders(request):

            placeholder = placeholder_cls(self)
            placeholder.request = request
            placeholder.workspace = workspace
            if entries:
                placeholder.dashboard_entries = entries.get(
                    placeholder_cls.uid, None
                )

            placeholder_instances.append(placeholder)

        return placeholder_instances

    @property
    def primary_html_class(self):
        return 'layout-{0}'.format(self.uid)

    @property
    def html_class(self):
        """
        Class used in the HTML.

        :return string:
        """
        return '{0} {1}'.format(self.primary_html_class,
                                ' '.join(self.html_classes))

    def get_css(self, placeholders):
        """Get placeholder specific css.

        :param iterable placeholders: Iterable of
            ``dash.base.BaseDashboardPlaceholder`` subclassed instances.
        :return string:
        """
        css = []
        for placeholder in placeholders:
            css.append(placeholder.css)
        return '\n'.join(css)

    def render_for_view(self, dashboard_entries=None, workspace=None,
                        request=None):
        """Render the layout.

        NOTE: This is not used at the moment. You most likely want the
        ``dash.views.dashboard`` view.

        :param iterable dashboard_entries:
        :param string workspace: Current workspace.
        :param django.http.HttpRequest request:
        :return string:
        """
        placeholders = self.get_placeholder_instances(
            dashboard_entries, workspace, request
        )
        context = {
            'placeholders': placeholders,
            'placeholders_dict': iterable_to_dict(placeholders,
                                                  key_attr_name='uid'),
            'request': request,
            'css': self.get_css(placeholders)
        }
        return render_to_string(self.get_view_template_name(request), context)

    def render_for_edit(self, dashboard_entries=None, workspace=None,
                        request=None):
        """Render the layout.

        NOTE: This is not used at the moment. You most likely want the
        ``dash.views.edit_dashboard`` view.

        :param iterable dashboard_entries:
        :param string workspace: Current workspace.
        :param django.http.HttpRequest request:
        :return string:
        """
        placeholders = self.get_placeholder_instances(
            dashboard_entries, workspace, request
        )
        context = {
            'placeholders': placeholders,
            'placeholders_dict': iterable_to_dict(placeholders,
                                                  key_attr_name='uid'),
            'request': request,
            'css':  self.get_css(placeholders)
        }
        return render_to_string(self.get_edit_template_name(request), context)


class BaseDashboardPlaceholder(object):
    """Base placeholder.

    :Properties:
        - `uid` (string): Unique identifier (shouldn't repeat within a single
          layout).
        - `cols` (int): Number of cols in the placeholder.
        - `rows` (int): Number of rows in the placeholder.
        - `cell_width` (int): Single cell (1x1) width.
        - `cell_height` (int): Single cell (1x1) height.
        - `cell_margin_top` (int): Top margin of a single cell.
        - `cell_margin_right` (int): Right margin of a single cell.
        - `cell_margin_bottom` (int): Bottom margin of a single cell.
        - `cell_margin_left` (int): Left margin of a single cell.
        - `view_template_name` (string): Template to be used for rendering the
          placeholder in view mode.
        - `edit_template_name` (string): Template to be used for rendering the
          placeholder in edit mode.
        - `html_classes` (string): Extra HTML class that layout should get.
    """

    uid = None
    cols = None  # 1
    rows = None  # 8
    cell_width = None  # 100
    cell_height = None  # 100
    cell_margin_top = 0
    cell_margin_right = 0
    cell_margin_bottom = 0
    cell_margin_left = 0
    view_template_name = ''
    edit_template_name = ''
    html_classes = []

    def __init__(self, layout):
        assert self.uid
        assert self.rows
        assert self.cols
        assert self.cell_width
        assert self.cell_height

        self.layout = layout
        self.dashboard_entries = None
        self.request = None

    @property
    def cell_units(self):
        """Cell units."""
        return self.layout.cell_units

    @property
    def html_id(self):
        """ID (unique) used in the HTML.

        :return string:
        """
        return 'id_{0}'.format(self.uid)

    @property
    def primary_html_class(self):
        """Primary HTML class."""
        return 'placeholder-{0}'.format(self.uid)

    @property
    def html_class(self):
        """Class used in the HTML.

        :return string:
        """
        return '{0} {1}'.format(self.primary_html_class,
                                ' '.join(self.html_classes))

    def get_view_template_name(self):
        return self.view_template_name \
            if self.view_template_name \
            else DEFAULT_PLACEHOLDER_VIEW_TEMPLATE_NAME

    def get_edit_template_name(self):
        return self.edit_template_name \
            if self.edit_template_name \
            else DEFAULT_PLACEHOLDER_EDIT_TEMPLATE_NAME

    def load_dashboard_entries(self, dashboard_entries=None):
        """Feed the dashboard entries to the layout for rendering later.

        :param iterable dashboard_entries: Iterable of
            ``dash.models.DashboardEntry`` objects.
        """
        self.dashboard_entries = dashboard_entries

    def render_for_view(self):
        """Render the placeholder for view mode.

        :return string:
        """
        context = {
            'placeholder': self,
            'dashboard_entries': self.dashboard_entries,
            'request': self.request,
            'workspace': self.workspace
        }
        return render_to_string(self.get_view_template_name(), context)

    def _generate_widget_cells(self):
        """Generates widget cells.

        Return a list of tuples, where the first element represents the cell
        class and the second element represents the cell position.

        :param string workspace: Current workspace slug.
        :param django.http.HttpRequest request:
        :return list:
        """
        empty_cells = []
        position = 1
        for row in range(1, self.rows + 1):
            for col in range(1, self.cols + 1):
                empty_cells.append(
                    ('col-{0} row-{1}'.format(col, row), position)
                )
                position += 1
        return empty_cells

    def render_for_edit(self):
        """Render the placeholder for edit mode.

        :return string:
        """
        context = {
            'placeholder': self,
            'dashboard_entries': self.dashboard_entries,
            'request': self.request,
            'workspace': self.workspace,
            'widget_cells': self._generate_widget_cells()
        }
        return render_to_string(self.get_edit_template_name(), context)

    def get_cell_width(self):
        """Get a single cell width, with respect to margins.

        :return int:
        """
        return self.cell_margin_left + \
            self.cell_margin_right + \
            self.cell_width

    def get_cell_height(self):
        """Get a single cell height, with respect to margins.

        :return int:
        """
        return self.cell_margin_top + \
            self.cell_margin_bottom + \
            self.cell_height

    def widget_inner_width(self, cols):
        """The inner width of the widget to be rendered."""
        return (
            (self.get_cell_width() * cols) -
            self.cell_margin_left -
            self.cell_margin_right
        )

    def widget_inner_height(self, rows):
        """The inner height of the widget to be rendered.

        :return int:
        """
        return (
            (self.get_cell_height() * rows) -
            self.cell_margin_top -
            self.cell_margin_bottom
        )

    @property
    def css(self):
        """CSS styles for the placeholders and plugins.

        The placeholder dimensions as well as columns sizes, should be
        handled here. Since we are in a placeholder and a placeholder has a
        defined number of rows and columns and each render has just a fixed
        amount of rows and columns defined, we can render the top left
        corners generic css classes.

        Cells do NOT have margins or paddings. This is essential (since all
        the plugins are positioned absolutely). If you want to have padding in
        your plugin widget, specify the `plugin-content-wrapper` class style
        in your specific layout/theme.

        :example:

            .placeholder .plugin .plugin-content-wrapper {
                padding: 5px;
            }

        :return string:
        """
        def placeholder_width():
            """Placeholder width.

            :return string:
            """
            return '{0}{1}'.format(self.cols * self.get_cell_width(),
                                   self.cell_units)

        def placeholder_height():
            """Placeholder height.

            :return string:
            """
            return '{0}{1}'.format(self.rows * self.get_cell_height(),
                                   self.cell_units)

        def plugin_width():
            """Default width of a plugin widget (1 cell).

            :return string:
            """
            return '{0}{1}'.format(self.cell_width, self.cell_units)

        def plugin_height():
            """Default height of a plugin widget (1 cell).

            :return string:
            """
            return '{0}{1}'.format(self.cell_height, self.cell_units)

        def plugin_positions():
            """Plugin positions depending on the row and cell occupied.

            All plugins are positioned absolutely. Based on the row, we use
            `margin-top` and `margin-left` to position a plugin.

            ..:Used CSS classes:
                - `row-1`, `row-2`, etc.
                - `col-1`, `col-2`, etc.

            :return string:
            """
            positions = []
            row_template = """
            .placeholder.{placeholder_class} .empty-widget-cell.row-{row_num},
            .placeholder.{placeholder_class} .plugin.row-{row_num} {{
                margin-top: {top};
            }}
            """
            for row_num in range(0, self.rows):
                _val = row_template.format(
                    placeholder_class=self.primary_html_class,
                    row_num=(row_num + 1),
                    top='{0}{1}'.format(self.get_cell_height() * row_num,
                                        self.cell_units)
                )
                positions.append(_val)

            col_template = """
            .placeholder.{placeholder_class} .empty-widget-cell.col-{col_num},
            .placeholder.{placeholder_class} .plugin.col-{col_num} {{
                margin-left: {left};
            }}
            """
            for col_num in range(0, self.cols):
                _val = col_template.format(
                    placeholder_class=self.primary_html_class,
                    col_num=(col_num + 1),
                    left='{0}{1}'.format(self.get_cell_width() * col_num,
                                         self.cell_units)
                )
                positions.append(_val)

            return '\n'.join(positions)

        def plugin_sizes():
            """Plugin size based on its' `rows` and `cols` properties.

            ..:Used CSS classes:
                - `width-1`, `width-2`, etc.
                - `height-1`, `height-2`, etc.

            :return string:
            """
            sizes = []

            row_template = """
            .placeholder.{placeholder_class} .plugin.height-{row_num} {{
                height: {height};
            }}
            """
            for row_num in range(0, self.rows):
                _val = row_template.format(
                    placeholder_class=self.primary_html_class,
                    row_num=(row_num + 1),
                    height='{0}{1}'.format(
                        self.widget_inner_height(row_num + 1),
                        self.cell_units
                    )
                )
                sizes.append(_val)

            col_template = """
            .placeholder.{placeholder_class} .plugin.width-{col_num} {{
                width: {width};
            }}
            """
            for col_num in range(0, self.cols):
                _val = col_template.format(
                    placeholder_class=self.primary_html_class,
                    col_num=(col_num + 1),
                    width='{0}{1}'.format(
                        self.widget_inner_width(col_num + 1),
                        self.cell_units
                    )
                )
                sizes.append(_val)

            return '\n'.join(sizes)

        def empty_cell_size():
            """CSS for empty cell size."""
            _val = """
            .placeholder.{placeholder_class} .empty-widget-cell {{
                width: {width};
                height: {height};
                line-height: {height};
            }}
            """.format(
                placeholder_class=self.primary_html_class,
                width='{0}{1}'.format(self.cell_width, self.cell_units),
                height='{0}{1}'.format(self.cell_height, self.cell_units)
            )
            return _val

        css = """
        .placeholder.{placeholder_class} {{
            width: {placeholder_width};
            height: {placeholder_height};
        }}

        .placeholder.{placeholder_class} .plugin {{
            width: {plugin_width};
            height: {plugin_height};
        }}

        {plugin_positions}

        {plugin_sizes}

        {empty_cell_sizes}
        """.format(
            placeholder_class=self.primary_html_class,
            placeholder_width=placeholder_width(),
            placeholder_height=placeholder_height(),
            plugin_width=plugin_width(),
            plugin_height=plugin_height(),
            plugin_positions=plugin_positions(),
            plugin_sizes=plugin_sizes(),
            empty_cell_sizes=empty_cell_size()
        )

        return css


class DashboardPluginDataStorage(object):
    """Storage for plugin data."""


class BaseDashboardPlugin(object):
    """Base dashboard plugin from which every plugin should inherit.

    :Properties:
        - `uid` (string): Plugin uid (obligatory). Example value: 'dummy',
          'wysiwyg', 'news'.
        - `name` (string): Plugin name (obligatory). Example value:
          'Dummy plugin', 'WYSIWYG', 'Latest news'.
        - `description` (string): Plugin description (optional). Example
          value: 'Dummy plugin used just for testing'.
        - `help_text` (string): Plugin help text (optional). This text would
          be shown in ``dash.views.add_dashboard_entry``.
          and ``dash.views.edit_dashboard_entry`` views.
        - `form`: Plugin form (optional). A subclass of ``django.forms.Form``.
          Should be given in case plugin is configurable.
        - `add_form_template` (str) (optional): Add form template (optional).
          If given, overrides the ``dash.views.add_dashboard_entry`` default
          template.
        - `edit_form_template` (string): Edit form template (optional). If
          given, overrides the ``dash.views.edit_dashboard_entry`` default
          template.
        - `html_classes` (list): List of extra HTML classes for the plugin.
        - `group` (string): Plugin are grouped under the specified group.
          Override in your plugin if necessary.
    """
    uid = None
    name = None
    description = None
    help_text = None
    form = None
    add_form_template = None
    edit_form_template = None
    html_classes = []
    group = _("General")

    def __init__(self, layout_uid, placeholder_uid, workspace=None,
                 user=None, position=None):
        """
        :param string placeholder_uid: Unique identifier of plugin
            placeholder (layout.placeholder).
        :param dash.models.DashboardWorkspace workspace: Plugin workspace.
        :param django.contrib.auth.models.User user: Plugin owner.
        """
        # Making sure all necessary properties are defined.
        try:
            assert self.uid
            assert self.name
        except Exception as e:
            raise NotImplementedError(
                "You should define `uid` and `name` properties in your "
                "`{0}.{1}` class.".format(self.__class__.__module__,
                                          self.__class__.__name__)
            )

        layout_cls = layout_registry.get(layout_uid, None)
        self.layout = layout_cls() if layout_cls else None

        placeholder_cls = self.layout.get_placeholder(placeholder_uid)
        self.placeholder = placeholder_cls(self.layout) \
            if placeholder_cls \
            else None

        if not (self.layout and self.placeholder):
            raise Exception(
                "Invalid placeholder value {0} in "
                "your `{1}.{2}` class.".format(
                    placeholder_uid,
                    self.__class__.__module__,
                    self.__class__.__name__
                )
            )

        self.layout_uid = layout_uid
        self.placeholder_uid = placeholder_uid
        self.workspace = workspace
        self.user = user
        self.position = position

        # Some initial values
        self.request = None

        self.data = DashboardPluginDataStorage()

        self._html_id = 'p{0}'.format(uuid.uuid4())

    @property
    def html_id(self):
        """HTML id."""
        return self._html_id

    def get_position(self):
        """Get the exact position of the plugin widget in the placeholder (row
        number, col number).

        :return tuple: Tuple of row and col numbers.
        """
        col = self.position % self.placeholder.cols
        row = int(
            self.position / self.placeholder.cols
        ) + (1 if col > 0 else 0)
        if col == 0:
            col = self.placeholder.cols

        return row, col

    @property  # Comment the @property out if something goes wrong.
    def html_class(self):
        """HTML class.

        A massive work on positioning the plugin and having it to be displayed
        in a given width is done here. We should be getting the plugin widget
        for the plugin given and based on its' properties (static!) as well as
        on plugin position (which we have from model), we can show the plugin
        with the exact class.
        """
        try:
            widget = self.get_widget()

            html_class = [
                'plugin-{0} {1} {2}'.format(
                    self.uid,
                    widget.html_class,
                    ' '.join(self.html_classes)
                )
            ]

            html_class.append('width-{0}'.format(widget.cols))
            html_class.append('height-{0}'.format(widget.rows))

            row, col = self.get_position()
            html_class.append('row-{0}'.format(row))
            html_class.append('col-{0}'.format(col))

            return ' '.join(html_class)
        except Exception as err:
            logger.debug(str(err))

    def process(self, plugin_data=None, fetch_related_data=False):
        """Init plugin with data."""
        try:
            # Calling pre-processor.
            self.pre_processor()

            if plugin_data:
                try:
                    # Trying to load the plugin data to JSON.
                    plugin_data = json.loads(plugin_data)

                    # If a valid JSON object, feed it to our plugin and
                    # process the data. The ``process_data`` method should
                    # be defined in your subclassed plugin class.
                    if plugin_data:
                        self.load_plugin_data(plugin_data)

                        self.process_plugin_data(
                            fetch_related_data=fetch_related_data
                        )
                except Exception as err:
                    if DEBUG:
                        logger.debug(str(err))

            # Calling the post processor.
            self.post_processor()

            return self
        except Exception as err:
            if DEBUG:
                logger.debug(str(err))

    def load_plugin_data(self, plugin_data):
        """Load the plugin data saved in ``dash.models.DashboardEntry``.
        Plugin data is saved in JSON string.

        :param string plugin_data: JSON string with plugin data.
        """
        self.plugin_data = plugin_data

    def _process_plugin_data(self, fields, fetch_related_data=False):
        """Process the plugin data.

        Override if need customisations.

        Beware, this is not always called.
        """
        for field, default_value in fields:
            try:
                setattr(
                    self.data,
                    field,
                    self.plugin_data.get(field, default_value)
                )
            except Exception:
                setattr(self.data, field, default_value)

    def process_plugin_data(self, fetch_related_data=False):
        """Processes the plugin data."""
        form = self.get_form()

        return self._process_plugin_data(
            form.plugin_data_fields,
            fetch_related_data=fetch_related_data
        )

    def _get_plugin_form_data(self, fields):
        """Get plugin data.

        :param iterable fields: List of tuples to iterate.
        :return dict:
        """
        form_data = {}
        for field, default_value in fields:
            try:
                form_data.update(
                    {field: self.plugin_data.get(field, default_value)}
                )
            except Exception as err:
                if DEBUG:
                    logger.debug(err)
        return form_data

    def get_plugin_form_data(self):
        """Get plugin form data.

        Fed as ``initial`` argument to the plugin form when initialising the
        instance for adding or editing the plugin. Override in your plugin
        class if you need customisations.
        """
        form = self.get_form()

        return self._get_plugin_form_data(form.plugin_data_fields)

    def get_instance(self):
        """Get instances."""
        return None

    def get_form(self):
        """Get the plugin form class.

        Override this method in your subclassed ``dash.base.DashboardPlugin``
        class when you need your plugin setup to vary depending on the
        placeholder, workspace, user or request given. By default returns the
        value of the ``form`` attribute defined in your plugin.

        :return django.forms.Form|django.forms.ModelForm: Subclass of
            ``django.forms.Form`` or ``django.forms.ModelForm``.
        """
        return self.form

    def get_initialised_create_form(self, data=None, files=None):
        """
        Used ``dash.views.add_dashboard_entry`` view to gets initialised form
        for object to be created.
        """
        plugin_form = self.get_form()
        if plugin_form:
            try:
                plugin_form = self.get_form()
                if plugin_form:
                    return plugin_form(data=data, files=files)
            except Exception as err:
                if DEBUG:
                    logger.debug(err)
                raise Http404(err)

    def get_initialised_create_form_or_404(self, data=None, files=None):
        """Get initialised create form or 404.

        Same as ``get_initialised_create_form`` but raises
        ``django.http.Http404`` on errors.
        """
        plugin_form = self.get_form()
        if plugin_form:
            try:
                return self.get_initialised_create_form(data=data, files=files)
            except Exception as e:
                if DEBUG:
                    logger.debug(e)
                raise Http404(e)

    def get_initialised_edit_form(self, data=None, files=None,
                                  auto_id='id_%s', prefix=None, initial=None,
                                  error_class=ErrorList, label_suffix=':',
                                  empty_permitted=False, instance=None):
        """Get initialised edit form.

        Used in ``dash.views.edit_dashboard_entry`` view.
        """
        plugin_form = self.get_form()
        if plugin_form:
            kwargs = {
                'data': data,
                'files': files,
                'auto_id': auto_id,
                'prefix': prefix,
                'initial': initial,
                'error_class': error_class,
                'label_suffix': label_suffix,
                'empty_permitted': empty_permitted
            }
            if issubclass(plugin_form, ModelForm):
                kwargs.update({'instance': instance})
            return plugin_form(**kwargs)

    def get_initialised_edit_form_or_404(self, data=None, files=None,
                                         auto_id='id_%s', prefix=None,
                                         error_class=ErrorList,
                                         label_suffix=':',
                                         empty_permitted=False):
        """Get initialised edit form or 404.

        Same as ``get_initialised_edit_form`` but raises
        ``django.http.Http404`` on errors.
        """
        plugin_form = self.get_form()
        if plugin_form:
            try:
                return self.get_initialised_edit_form(
                    data=data,
                    files=files,
                    auto_id=auto_id,
                    prefix=prefix,
                    initial=self.get_plugin_form_data(),
                    error_class=error_class,
                    label_suffix=label_suffix,
                    empty_permitted=empty_permitted,
                    instance=self.get_instance()
                )
            except Exception as err:
                if DEBUG:
                    logger.debug(err)
                raise Http404(err)

    def get_widget(self, request=None, as_instance=False):
        """Get the plugin widget.

        :param django.http.HttpRequest request:
        :param bool as_instance:
        :return mixed: Subclass of ``dash.base.BaseDashboardPluginWidget`` or
            instance of subclassed ``dash.base.BaseDashboardPluginWidget``
            object.
        """
        widget_cls = plugin_widget_registry.get(
            PluginWidgetRegistry.namify(self.layout.uid,
                                        self.placeholder.uid,
                                        self.uid)
            )

        if not as_instance:
            return widget_cls
        elif widget_cls:
            widget = widget_cls(self)
            return widget

    def render(self, request=None):
        """Render the plugin HTML (for dashboard workspace).

        :param django.http.HttpRequest request:
        :return string:
        """
        widget_cls = self.get_widget()

        if widget_cls:
            widget = widget_cls(self)

            render = widget.render(request=request)
            return render or ''
        elif DEBUG:
            logger.debug(
                "No widget defined for {0}.{1}.{2}".format(
                    self.layout.uid,
                    self.placeholder.uid,
                    self.uid
                )
            )

    def _update_plugin_data(self, dashboard_entry):
        """Update plugin data.

        For private use. Do not override this method. Override
        ``update_plugin_data`` instead.
        """
        try:
            updated_plugin_data = self.update_plugin_data(dashboard_entry)
            plugin_data = self.get_updated_plugin_data(
                update=updated_plugin_data
            )
            return self.save_plugin_data(
                dashboard_entry,
                plugin_data=plugin_data
            )
        except Exception as err:
            logging.debug(str(err))

    def update_plugin_data(self, dashboard_entry):
        """Update plugin data.

        Used in ``dash.management.commands.dash_update_plugin_data``.

        Some plugins would contain data fetched from various sources (models,
        remote data). Since dashboard entries are by definition loaded
        extremely much, you are advised to store as much data as possible in
        ``plugin_data`` field of ``dash.models.DashboardEntry``. Some
        externally fetched data becomes invalid after some time and needs
        updating. For that purpose, in case if your plugin needs that, redefine
        this method in your plugin. If you need your data to be periodically
        updated, add a cron-job which would run ``dash_update_plugin_data``
        management command (see
        ``dash.management.commands.dash_update_plugin_data`` module).

        :param dash.models.DashboardEntry dashboard_entry: Instance of
            ``dash.models.DashboardEntry``.
        :return dict: Should return a dictionary containing data of fields to
            be updated.
        """

    def _delete_plugin_data(self):
        """
        For private use. Do not override this method. Override
        ``delete_plugin_data`` instead.
        """
        try:
            self.delete_plugin_data()
        except Exception as e:
            logging.debug(str(e))

    def delete_plugin_data(self):
        """Delete plugin data.

        Used in ``dash.views.delete_dashboard_entry``. Fired automatically,
        when ``dash.models.DashboardEntry`` object is about to be deleted. Make
        use of it if your plugin creates database records or files that are
        not monitored externally but by dash only.
        """

    def _clone_plugin_data(self, dashboard_entry):
        """Clone plugin data.

        For private use. Do not override this method. Override
        ``clone_plugin_data`` instead.
        """
        try:
            return self.clone_plugin_data(dashboard_entry)
        except Exception as err:
            logging.debug(str(err))

    def clone_plugin_data(self, dashboard_entry):
        """Clone plugin data.

        Used when copying entries. If any objects or files are created by
        plugin, they should be cloned.

        :param dash.models.DashboardEntry dashboard_entry: Instance of
            ``dash.models.DashboardEntry``.
        :return string: JSON dumped string of the cloned plugin data. The
            returned value would be inserted as is into the
            ``dash.models.DashboardEntry.plugin_data`` field.
        """

    def get_cloned_plugin_data(self, update={}):
        """Get the cloned plugin data and returns it in a JSON dumped format.

        :param dict update:
        :return string: JSON dumped string of the cloned plugin data.

        :example:

        In the ``get_cloned_plugin_data`` method of your plugin, do as follows:

        >>> def clone_plugin_data(self, dashboard_entry):
        >>>     cloned_image = clone_file(self.data.image, relative_path=True)
        >>>     return self.get_cloned_plugin_data(
        >>>         update={'image': cloned_image}
        >>>     )
        """
        form = self.get_form()

        cloned_data = copy.copy(self.data)
        data = {}

        for field, default_value in form.plugin_data_fields:
            data.update({field: getattr(cloned_data, field, '')})

        for prop, value in update.items():
            data.update({prop: value})

        return json.dumps(data)

    def get_updated_plugin_data(self, update={}):
        """Get the plugin data and returns it in a JSON dumped format.

        :param dict update:
        :return string: JSON dumped string of the cloned plugin data.
        """
        form = self.get_form()
        data = {}

        for field, default_value in form.plugin_data_fields:
            data.update({field: getattr(self.data, field, '')})

        for prop, value in update.items():
            data.update({prop: value})

        return json.dumps(data)

    def pre_processor(self):
        """Pre-process data.

        Redefine in your subclassed plugin when necessary.

        Pre process plugin data (before rendering). This method is being
        called before the data has been loaded into the plugin.

        Note, that request (django.http.HttpRequest) is
        available (self.request).
        """

    def post_processor(self):
        """Post-process data.

        Redefine in your subclassed plugin when necessary.

        Post process plugin data here (before rendering). This method is being
        called after the data has been loaded into the plugin.

        Note, that request (django.http.HttpRequest) is
        available (self.request).
        """

    def save_plugin_data(self, dashboard_entry, plugin_data):
        """Save plugin data.

        Used in bulk update plugin data.

        :param dash.models.DashboardEntry dashboard_entry:
        :param dict plugin_data:
        :return bool: True if all went well.
        """
        try:
            if plugin_data:
                dashboard_entry.plugin_data = plugin_data
                dashboard_entry.save()
                return True
        except Exception as err:
            logger.debug(str(err))


class MetaBaseDashboardPluginWidget(type):
    """Meta class for ``dash.base.BaseDashboardPluginWidget``."""

    @property
    def html_class(cls):
        """HTML class of the ``dash.base.BaseDashboardPluginWidget``.

        :return string:
        """
        return ' '.join(cls.html_classes)


class ClassProperty(property):
    """Class property."""
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()


classproperty = ClassProperty


class BaseDashboardPluginWidget(object):
    """Base plugin widget.

    So, if we would want to register a plugin widget (renderer) for some
    layout, we would first define the plugin widget and then just write:

        >>> plugin_widget_registry.register(DummyPluginWidget)

    Plugin widget is always being registered for a placeholder. Placeholder in
    its' turn has number of rows and columns. Since we register each widget
    for a (layout, placeholder, plugin) combination separately, it fits the
    needs and requirements perfectly. In that way we are able to tell, whether
    plugin has a widget available and actually valid (qua dimensions) for the
    placeholder. Plugin is just data. Nothing more. Widget operates with that
    data. Thus, widget has number of rows and columns it occupies in the
    placeholder registered. By default, number of rows and columns is set to
    1, which means that a plugin occupies just 1 cell. But, certainly, there
    can be plugins that occupy more space in a placeholder.
    """

    layout_uid = None
    placeholder_uid = None
    plugin_uid = None
    cols = 1
    rows = 1
    html_classes = []
    media_js = []
    media_css = []

    def __init__(self, plugin):
        assert self.layout_uid and self.layout_uid == plugin.layout.uid
        assert (
            self.placeholder_uid
            and
            self.placeholder_uid in plugin.layout.placeholder_uids
        )
        assert (
            self.plugin_uid
            and
            self.plugin_uid in get_registered_plugin_uids()
        )
        assert hasattr(self, 'render') and callable(self.render)
        assert self.cols
        assert self.rows
        assert isinstance(self.media_js, (list, tuple))
        assert isinstance(self.media_css, (list, tuple))

        if isinstance(self.media_js, tuple):
            self.media_js = list(self.media_js)

        if isinstance(self.media_css, tuple):
            self.media_css = list(self.media_css)

        self.plugin = plugin

    def render(self, request=None):
        """Render.

        :param django.http.HttpRequest request:
        :return str:
        """
        return ''

    @classproperty
    def html_class(cls):
        """HTML class of the ``dash.base.BaseDashboardPluginWidget``.

        :return string:
        """
        return ' '.join(cls.html_classes)

    def get_width(self):
        """Get widget width.

        :return int:
        """
        return self.plugin.placeholder.widget_inner_width(self.cols)

    def get_height(self):
        """Get widget height.

        :return int:
        """
        return self.plugin.placeholder.widget_inner_height(self.rows)

    def get_size(self, delta_width=0, delta_height=0):
        """Get widget size.

        :param int delta_width:
        :param int delta_height:
        :return tuple:
        """
        return (
            (
                self.cols * self.plugin.placeholder.get_cell_width()
            ) + delta_width,
            (
                self.rows * self.plugin.placeholder.get_cell_height()
            ) + delta_height
        )


class BaseRegistry(object):
    """Registry of dash plugins.

    It's essential, that class registered has the``uid`` property.
    """

    type = None

    def __init__(self):
        assert self.type
        self._registry = {}
        self._forced = []

    def register(self, cls, force=False):
        """
        Registers the plugin in the registry.

        :param cls mixed.
        :param bool force:
        """
        if not issubclass(cls, self.type):
            raise InvalidRegistryItemType(
                "Invalid item type `{0}` for "
                "registry `{1}`".format(cls, self.__class__)
            )

        # If item has not been forced yet, add/replace its' value in the
        # registry
        if force:

            if cls.uid not in self._forced:
                self._registry[cls.uid] = cls
                self._forced.append(cls.uid)
                return True
            else:
                return False

        else:

            if cls.uid in self._registry:
                return False
            else:
                self._registry[cls.uid] = cls
                return True

    def unregister(self, cls):
        if not issubclass(cls, self.type):
            raise InvalidRegistryItemType(
                "Invalid item type `{0}` for "
                "registry `{1}`".format(cls, self.__class__)
            )

        # Only non-forced items are allowed to be unregistered.
        if cls.uid in self._registry and cls.uid not in self._forced:
            self._registry.pop(cls.uid)
            return True
        else:
            return False

    def get(self, uid, default=None):
        """Get the given entry from the registry.

        :param string uid:
        :param mixed default:
        :return mixed.
        """
        item = self._registry.get(uid, default)
        if not item:
            logger.debug(
                "Can't find plugin with uid `{0}` in `{1}` "
                "registry".format(uid, self.__class__)
            )
        return item


class PluginRegistry(BaseRegistry):
    """Plugin registry."""

    type = BaseDashboardPlugin


class LayoutRegistry(BaseRegistry):
    """Layout registry."""

    type = BaseDashboardLayout


class PluginWidgetRegistry(object):
    """Registry of dash plugins widgets (renderers)."""

    type = BaseDashboardPluginWidget

    def __init__(self):
        assert self.type
        self._registry = {}
        self._forced = []

    @staticmethod
    def namify(layout, placeholder, plugin_uid):
        return '{0}.{1}.{2}'.format(layout, placeholder, plugin_uid)

    def register(self, cls, force=False):
        """Register the plugin renderer in the registry.

        :param dash.base.BasePluginRenderer cls: Subclass of
            ``dash.base.BasePluginRenderer``.
        """
        if not issubclass(cls, self.type):
            raise InvalidRegistryItemType(
                "Invalid item type `{0}` for "
                "registry `{1}`".format(cls, self.__class__)
            )

        uid = PluginWidgetRegistry.namify(
            cls.layout_uid,
            cls.placeholder_uid,
            cls.plugin_uid
        )

        # If item has not been forced yet, add/replace its' value in the
        # registry.
        if force:

            if uid not in self._forced:
                self._registry[uid] = cls
                self._forced.append(uid)
                return True
            else:
                return False

        else:

            if uid in self._registry:
                return False
            else:
                self._registry[uid] = cls
                return True

    def unregister(self, cls):
        if not issubclass(cls, self.type):
            raise InvalidRegistryItemType(
                "Invalid item type `{0}` for "
                "registry `{1}`".format(cls, self.__class__)
            )

        uid = PluginWidgetRegistry.namify(
            cls.layout_uid,
            cls.placeholder_uid,
            cls.plugin_uid
        )

        # Only non-forced items are allowed to be unregistered.
        if uid in self._registry and uid not in self._forced:
            self._registry.pop(uid)
            return True
        else:
            return False

    def get(self, uid, default=None):
        """Get the given entry from the registry.

        :param string uid:
        :param mixed default:
        :return mixed.
        """
        item = self._registry.get(uid, default)
        if not item:
            logger.debug(
                "Can't find plugin with uid `{0}` in `{1}` "
                "registry".format(uid, self.__class__)
            )
        return item


# Register plugins by calling plugin_registry.register()
plugin_registry = PluginRegistry()

# Register layouts by calling layout_registry.register()
layout_registry = LayoutRegistry()

# Register of plugin widgets.
plugin_widget_registry = PluginWidgetRegistry()


def ensure_autodiscover():
    """Ensure that plugins are auto-discovered."""
    if not (
        plugin_registry._registry and
        layout_registry._registry and
        plugin_widget_registry._registry
    ):
        autodiscover()


def get_registered_plugins():
    """Get a list of registered plugins in a form if tuple.

    Get a list of registered plugins in a form if tuple (plugin name, plugin
    description). If not yet auto-discovered, auto-discovers them.

    :return list:
    """
    ensure_autodiscover()

    registered_plugins = []

    for uid, plugin in plugin_registry._registry.items():
        registered_plugins.append((uid, safe_text(plugin.name)))

    return registered_plugins


def get_registered_plugin_uids():
    """Gets a list of registered plugin uids as a list .

    If not yet auto-discovered, auto-discovers them.

    :return list:
    """
    ensure_autodiscover()

    registered_plugins = []

    for uid, plugin in plugin_registry._registry.items():
        registered_plugins.append(uid)
    return registered_plugins


def validate_placeholder_uid(layout, placeholder_uid):
    """Validate the placeholder.

    :param string layout:
    :param string placeholder_uid:
    :return bool:
    """
    return placeholder_uid in layout.placeholder_uids


def validate_plugin_uid(plugin_uid):
    """Validate the plugin uid.

    :param string plugin_uid:
    :return bool:
    """
    return plugin_uid in get_registered_plugin_uids()


def get_registered_layouts():
    """Get registered layouts."""
    ensure_autodiscover()

    registered_layouts = []

    for uid, layout in layout_registry._registry.items():
        registered_layouts.append((uid, safe_text(layout.name)))

    return registered_layouts


def get_registered_layout_uids():
    """Get uids of registered layouts."""
    return layout_registry._registry.keys()


def get_layout(layout_uid=None, as_instance=False):
    """Gets the layout by ``layout_uid`` given.

    If left empty, takes the default one chosen in settings.

    Raises a ``dash.exceptions.NoActiveLayoutChosen`` when no default layout
    could be found.

    :return dash.base.BaseDashboardLayout: Sublcass of
        ``dash.base.BaseDashboardLayout``.
    """
    ensure_autodiscover()

    if not layout_uid:
        layout_uid = ACTIVE_LAYOUT

    layout_cls = layout_registry.get(layout_uid, None)
    if not layout_cls:
        raise LayoutDoesNotExist(
            _("Layout `{0}` does not exist!").format(layout_uid)
        )

    if as_instance:
        return layout_cls()

    return layout_cls


def collect_widget_media(dashboard_entries):
    """Collect the widget media for dashboard entries given.

    :param iterable dashboard_entries: Iterable of
        ``dash.models.DashboardEntry`` instances.
    :return dict: Returns a dict containing the 'js' and 'css' keys.
        Correspondent values of those keys are lists containing paths to the
        CSS and JS media files.
    """
    media_js = []
    media_css = []
    for dashboard_entry in dashboard_entries:
        widget_cls = plugin_widget_registry.get(
            PluginWidgetRegistry.namify(
                dashboard_entry.layout_uid,
                dashboard_entry.placeholder_uid,
                dashboard_entry.plugin_uid
            )
        )
        if widget_cls:
            media_js += widget_cls.media_js
            media_css += widget_cls.media_css
        else:
            logger.debug(
                "widget_cls empty for dashboard "
                "entry {0}".format(dashboard_entry.__dict__)
            )
    return {'js': media_js, 'css': media_css}



__title__ = 'dash.contrib.apps.public_dashboard'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = ('default_app_config',)

default_app_config = 'dash.contrib.apps.public_dashboard.apps.Config'



from dash.contrib.plugins.weather.conf import get_setting

__title__ = 'dash.contrib.plugins.weather.settings'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'API_ENDPOINT_URL',
    'API_KEY',
    'DEFAULT_CACHE_FOR',
    'DEFAULT_SHOW_TITLE',
)

API_KEY = get_setting('API_KEY')
API_ENDPOINT_URL = get_setting('API_ENDPOINT_URL')
DEFAULT_SHOW_TITLE = get_setting('DEFAULT_SHOW_TITLE')
DEFAULT_CACHE_FOR = get_setting('DEFAULT_CACHE_FOR')



from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import ugettext_lazy as _

from ....base import BaseDashboardPlugin
from ....factory import plugin_factory

from .forms import URLForm, BookmarkForm
from .models import Bookmark

__title__ = 'dash.contrib.plugins.image.dash_plugins'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = ('BaseURLPlugin',)

# ****************************************************************************
# ********************************* URL plugin *******************************
# ****************************************************************************


class BaseURLPlugin(BaseDashboardPlugin):
    """Base URL plugin."""

    name = _("URL")
    group = _("URLs")
    form = URLForm

    @property
    def html_class(self):
        """HTML class.

        If plugin has an image, we add a class ``iconic`` to it.
        """
        html_class = super(BaseURLPlugin, self).html_class
        if self.data.image:
            html_class += ' iconic-url'
        return html_class

# ****************************************************************************
# ********** Generating and registering the URL plugins using factory ********
# ****************************************************************************


sizes = (
    (1, 1),
    (2, 2)
)

plugin_factory(BaseURLPlugin, 'url', sizes)


# ****************************************************************************
# ********************************* Bookmark plugin **************************
# ****************************************************************************


class BaseBookmarkPlugin(BaseDashboardPlugin):
    """Base URL plugin."""

    name = _("Bookmark")
    group = _("URLs")
    form = BookmarkForm

    @property
    def html_class(self):
        """HTML class.

        If plugin has an image, we add a class `iconic` to it.
        """
        html_class = super(BaseBookmarkPlugin, self).html_class
        if self.data.image:
            html_class += ' iconic-url'
        return html_class

    def update_plugin_data(self, dashboard_entry):
        """Update plugin data.

        Should return a dictionary with the plugin data which is supposed to
        be updated.
        """
        try:
            bookmark = Bookmark._default_manager.get(pk=self.data.bookmark)
        except ObjectDoesNotExist:
            return

        if bookmark:
            data = {
                'bookmark': bookmark.pk,
                'title': bookmark.title,
                'url': bookmark.url,
                'external': bookmark.external,
                'image': bookmark.image
            }
            return data

# ****************************************************************************
# ******* Generating and registering the Bookmark plugins using factory ******
# ****************************************************************************


sizes = (
    (1, 1),
    # (2, 2)
)

plugin_factory(BaseBookmarkPlugin, 'bookmark', sizes)



__title__ = 'dash.contrib.plugins.dummy.defaults'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'DEFAULT_MAX_CHARS',
    'LANGUAGE_CHOICES',
    'LANGUAGE_CHOICES_KEYS',
)

LANGUAGE_CHOICES = (
    ('en', 'English'),
    ('hy', 'Armenian'),
    ('ka', 'Georgian'),
    ('el', 'Greek'),
    ('ru', 'Russian')
)

LANGUAGE_CHOICES_KEYS = [__val[0] for __val in LANGUAGE_CHOICES]

DEFAULT_MAX_CHARS = 40



from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import ugettext_lazy as _

from .base import validate_placeholder_uid
from .helpers import (
    clone_plugin_data,
    lists_overlap,
    safe_text
)
from .models import (
    DashboardEntry,
    DashboardWorkspace
)
from .settings import PLUGIN_CLIPBOARD_KEY
from .utils import (
    build_cells_matrix,
    get_occupied_cells,
    get_user_plugin_uids
)

__title__ = 'dash.clipboard'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'can_paste_from_clipboard',
    'copy_entry_to_clipboard',
    'cut_entry_to_clipboard',
    'get_plugin_data_from_clipboard',
    'paste_from_clipboard',
    'save_plugin_data_to_clipboard',
)


def make_session_key(layout_uid):
    """Make a session key based on the layout UID.

    :param str layout_uid:
    :return str:
    """
    return "{0}_{1}".format(PLUGIN_CLIPBOARD_KEY, layout_uid)


def save_plugin_data_to_clipboard(request,
                                  layout_uid,
                                  plugin_uid,
                                  plugin_data):
    """Save plugin to clipboard for later pasting.

    :param django.http.HttpRequest request:
    :param str layout_uid:
    :param str plugin_uid:
    :param str plugin_data: JSON data (stringified).
    :return bool: True on success. False otherwise.
    """
    request.session[make_session_key(layout_uid)] = {
        'plugin_uid': plugin_uid,
        'plugin_data': plugin_data
    }
    return True


def get_plugin_data_from_clipboard(request, layout_uid):
    """Retrieve the plugin clipboard.

    Note, that only last element is being saved.

    :param django.http.HttpRequest request:
    :param str layout_uid:
    :return dict:
    """
    session_key = make_session_key(layout_uid)
    if session_key in request.session:
        return request.session[session_key]


def clear_clipboard_data(request, layout_uid):
    """Clear the clipboard data."""
    session_key = make_session_key(layout_uid)
    if session_key in request.session:
        request.session[session_key] = None


def cut_entry_to_clipboard(request,
                           dashboard_entry,
                           delete_original_entry=True,
                           clone_data=True):
    """Cut ``dash.models.DashboardEntry`` to clipboard.

    :param django.http.HttpRequest request:
    :param dash.models.DashboardEntry dashboard_entry:
    :param bool delete_original_entry: If set to True, original entry is
        deleted; kept otherwise.
    :param bool clone_data: If set to Ture, plugin data is cloned.
    :return bool: True on success. False otherwise.
    """
    layout_uid = dashboard_entry.layout_uid
    plugin_uid = dashboard_entry.plugin_uid

    if clone_data:
        # plugin = dashboard_entry.get_plugin(request=request)

        # Getting cloned plugin data
        # plugin_data = plugin._clone_plugin_data(dashboard_entry)
        plugin_data = clone_plugin_data(dashboard_entry, request=request)

        # If none returned, using the still the current one (not cloned, which
        # is dangerous). Possible TODO for later.
        if plugin_data is None:
            plugin_data = dashboard_entry.plugin_data
    else:
        plugin_data = dashboard_entry.plugin_data

    save_plugin_data_to_clipboard(
        request=request,
        layout_uid=layout_uid,
        plugin_uid=plugin_uid,
        plugin_data=plugin_data
    )

    if delete_original_entry:
        dashboard_entry.delete()

    return True


def copy_entry_to_clipboard(request, dashboard_entry):
    """Copy ``dash.models.DashboardEntry`` to clipboard.

    :param django.http.HttpRequest request:
    :param dash.models.DashboardEntry dashboard_entry:
    :return bool: True on success. False otherwise.
    """
    return cut_entry_to_clipboard(
        request,
        dashboard_entry,
        delete_original_entry=False
    )


def paste_entry_from_clipboard(request,
                               layout,
                               placeholder_uid,
                               position,
                               workspace=None,
                               entries=[],
                               check_only=False,
                               clear_clipboard=True):
    """Paste entry from clipboard

    Paste entry from clipboard to the given placeholder of a workspace
    selected.

    :param django.http.HttpRequest request:
    :param str layout: Layout UID.
    :param str placeholder_uid: Placeholder UID.
    :param int position: Position.
    :param mixed workspace: Either str or ``dash.models.DashboardWorkspace``
        instance. If str is given, a database hit occurs in order to obtain
        the ``dash.models.DashboardWorkspace`` instance. Thus, if you have the
        workspace instance already, provide it as is, in order to minify
        database hits.
    :param iterable entries: If given, entries are not fetched but the
        existing iterable is used. Each item in the iterable should be an
        instance of ``dash.models.DashboardEntry``.
    :param bool check_only: If set to True, it's only checked if it's possible
        to paste from clipboard (the ``dashboard_entry.save()`` part is
        skipped, which means that the entry is not saved in the database).
    :return tuple (str, bool): (Plugin name, boolean True) tuple on success
        and (error message, boolean False) on failure.
    """
    clipboard_plugin_data = get_plugin_data_from_clipboard(request, layout.uid)
    if not clipboard_plugin_data:
        return _("Clipboard is empty!"), False

    if not validate_placeholder_uid(layout, placeholder_uid):
        return _("Invalid placeholder `{0}`.").format(placeholder_uid), False

    user_plugin_uids = get_user_plugin_uids(request.user)

    if isinstance(workspace, DashboardWorkspace):
        workspace_obj = workspace
    else:
        if workspace:
            try:
                workspace_obj = DashboardWorkspace._default_manager \
                                                  .get(slug=workspace)
            except ObjectDoesNotExist as e:
                workspace_obj = None
        else:
            workspace_obj = None

    dashboard_entry = DashboardEntry(
        user=request.user,
        workspace=workspace_obj,
        layout_uid=layout.uid,
        placeholder_uid=placeholder_uid,
        plugin_uid=clipboard_plugin_data['plugin_uid'],
        plugin_data=clipboard_plugin_data['plugin_data'],
        position=position
    )

    # We should now check if we have a space for pasting the plugin. For that
    # we should get the plugin and see if there's a space available for
    # the (workspace, placeholder, user) triple given.

    # Get the plugin.
    plugin = dashboard_entry.get_plugin()

    if plugin.uid not in user_plugin_uids:
        return (_("You're not allowed to "
                  "use the {0} plugin.".format(safe_text(plugin.name))), False)

    # Getting occupied cells
    placeholder = layout.get_placeholder(placeholder_uid)
    occupied_cells = build_cells_matrix(
        request.user,
        layout,
        placeholder,
        workspace=workspace
    )

    # Get cells occupied by plugin widget.
    widget_occupied_cells = get_occupied_cells(
        layout,
        placeholder,
        plugin.uid,
        position,
        check_boundaries=True
    )

    if widget_occupied_cells is not False \
       and not lists_overlap(widget_occupied_cells, occupied_cells):

        try:
            if not check_only:
                # Save data
                dashboard_entry.save()

                # Clear the clipboard
                clear_clipboard_data(request, layout.uid)

            return plugin.uid, True
        except Exception as err:
            return str(err), False


def can_paste_entry_from_clipboard(request,
                                   layout,
                                   placeholder_uid,
                                   position,
                                   workspace=None,
                                   entries=[]):
    """Check if clipboard plugin can be pasted into the placeholder.

    For actually pasting the plugin from clipboard into the placeholder, use
    ``dash.clipboard.paste_from_clipboard`` function.

    :param django.http.HttpRequest request:
    :param str layout:
    :param str placeholder_uid: Placeholder UID.
    :param int position:
    :param mixed workspace: Either str or ``dash.models.DashboardWorkspace``
        instance. If str is given, a database hit occurs in order to obtain
        the ``dash.models.DashboardWorkspace`` instance. Thus, if you have the
        workspace instance already, provide it as is, in order to minify
        database hits.
    :param iterable entries: If given, entries are not fetched but the existing
        iterable is used. Each item in the iterable should be an instance
        of ``dash.models.DashboardEntry``.
    :return tuple (str, bool): (Plugin name, boolean True) tuple on success
        and (error message, boolean False) on failure.
    """
    return paste_entry_from_clipboard(
        request=request,
        layout=layout,
        placeholder_uid=placeholder_uid,
        position=position,
        workspace=workspace,
        entries=entries,
        check_only=True
    )



from django.conf.urls import url
from django.utils.translation import ugettext_lazy as _

from .views import (
    add_dashboard_entry,
    clone_dashboard_workspace,
    copy_dashboard_entry,
    create_dashboard_workspace,
    cut_dashboard_entry,
    dashboard,
    dashboard_workspaces,
    delete_dashboard_entry,
    delete_dashboard_workspace,
    edit_dashboard,
    edit_dashboard_entry,
    edit_dashboard_settings,
    edit_dashboard_workspace,
    paste_dashboard_entry,
    plugin_widgets,
)

__title__ = 'dash.urls'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2013-2017 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = ('urlpatterns',)


urlpatterns = [
    # Paste dashboard entry
    url(_(r'^entry/paste/(?P<placeholder_uid>[\w_]+)/'
          r'(?P<workspace>[\w_\-]+)/pos/(?P<position>\d+)/$'),
        view=paste_dashboard_entry,
        name='dash.paste_dashboard_entry'),
    url(_(r'^entry/paste/(?P<placeholder_uid>[\w_]+)/pos/(?P<position>\d+)/$'),
        view=paste_dashboard_entry,
        name='dash.paste_dashboard_entry'),

    # Cut dashboard entry
    url(_(r'^entry/cut/(?P<entry_id>\d+)/$'),
        view=cut_dashboard_entry,
        name='dash.cut_dashboard_entry'),

    # Copy dashboard entry
    url(_(r'^entry/copy/(?P<entry_id>\d+)/$'),
        view=copy_dashboard_entry,
        name='dash.copy_dashboard_entry'),

    # Add dashboard entry.
    url(_(r'^entry/add/(?P<placeholder_uid>[\w_]+)/(?P<plugin_uid>[\w_\-]+)/'
          r'ws/(?P<workspace>[\w_\-]+)/pos/(?P<position>\d+)/$'),
        view=add_dashboard_entry,
        name='dash.add_dashboard_entry'),
    url(_(r'^entry/add/(?P<placeholder_uid>[\w_]+)/(?P<plugin_uid>[\w_\-]+)/'
          r'ws/(?P<workspace>[\w_\-]+)/$'),
        view=add_dashboard_entry,
        name='dash.add_dashboard_entry'),
    url(_(r'^entry/add/(?P<placeholder_uid>[\w_]+)/(?P<plugin_uid>[\w_\-]+)/'
          r'pos/(?P<position>\d+)/$'),
        view=add_dashboard_entry,
        name='dash.add_dashboard_entry'),
    url(_(r'^entry/add/(?P<placeholder_uid>[\w_]+)/'
          r'(?P<plugin_uid>[\w_\-]+)/$'),
        view=add_dashboard_entry,
        name='dash.add_dashboard_entry'),

    # Edit dashboard entry.
    url(_(r'^entry/edit/(?P<entry_id>\d+)/$'),
        view=edit_dashboard_entry,
        name='dash.edit_dashboard_entry'),

    # Delete dashboard entry.
    url(_(r'^entry/delete/(?P<entry_id>\d+)/$'),
        view=delete_dashboard_entry,
        name='dash.delete_dashboard_entry'),

    # ***************************************************************
    # ********************** Edit dashboard *************************
    # ***************************************************************
    # Edit dashboard.
    url(_(r'^edit/(?P<workspace>[\w_\-]+)/$'),
        view=edit_dashboard,
        name='dash.edit_dashboard'),
    url(_(r'^edit/$'),
        view=edit_dashboard,
        name='dash.edit_dashboard'),

    # ***************************************************************
    # ********************** Widgets for dashboard entries **********
    # ***************************************************************
    url(_(r'^plugin-widgets/(?P<placeholder_uid>[\w_]+)/'
          r'(?P<workspace>[\w_\-]+)/pos/(?P<position>\d+)/$'),
        view=plugin_widgets,
        name='dash.plugin_widgets'),
    # Workspace should not be named `pos`. Add check. TODO.
    url(_(r'^plugin-widgets/(?P<placeholder_uid>[\w_]+)/pos/'
          r'(?P<position>\d+)/$'),
        view=plugin_widgets,
        name='dash.plugin_widgets'),
    url(_(r'^plugin-widgets/(?P<placeholder_uid>[\w_]+)/'
          r'(?P<workspace>[\w_\-]+)/$'),
        view=plugin_widgets,
        name='dash.plugin_widgets'),
    url(_(r'^plugin-widgets/(?P<placeholder_uid>[\w_]+)/$'),
        view=plugin_widgets,
        name='dash.widgets'),

    # ***************************************************************
    # ********************** Dashboard workspace ********************
    # ***************************************************************
    # List workspaces.
    url(_(r'^workspaces/(?P<workspace>[\w_\-]+)/$'),
        view=dashboard_workspaces,
        name='dash.dashboard_workspaces'),
    url(_(r'^workspaces/$'),
        view=dashboard_workspaces,
        name='dash.dashboard_workspaces'),

    # Create dashboard workspace.
    url(_(r'^workspace/create/$'),
        view=create_dashboard_workspace,
        name='dash.create_dashboard_workspace'),

    # Edit dashboard workspace.
    url(_(r'^workspace/edit/(?P<workspace_id>\d+)/$'),
        view=edit_dashboard_workspace,
        name='dash.edit_dashboard_workspace'),

    # Delete dashboard workspace.
    url(_(r'^workspace/delete/(?P<workspace_id>\d+)/$'),
        view=delete_dashboard_workspace,
        name='dash.delete_dashboard_workspace'),

    # Clone dashboard workspace.
    url(_(r'^workspace/clone/(?P<workspace_id>\d+)/$'),
        view=clone_dashboard_workspace,
        name='dash.clone_dashboard_workspace'),

    # View dashboard workspace.
    url(_(r'^workspace/(?P<workspace>[\w_\-]+)/$'),
        view=dashboard,
        name='dash.dashboard'),

    # Edit dashboard settings.
    url(_(r'^settings/edit/$'),
        view=edit_dashboard_settings,
        name='dash.edit_dashboard_settings'),

    # View default dashboard (no workspace selected == default workspace used).
    url(_(r'^$'),
        view=dashboard,
        name='dash.dashboard'),
]
